from copy import deepcopy
import numpy as np
from posym.basis import BasisFunction, PrimitiveGaussian
from scipy.spatial.transform import Rotation as R
import itertools


def list_round(elements_list, decimals=2):
    """
    rounds a list of numbers and converts to Integer if necessary.
    Used for the string representation of symmetry objects

    :param elements_list: list of numbers
    :param decimals: decimal places used to round
    :return: list of rounded numbers
    """
    r_list = []
    for element in elements_list:
        if abs(np.round(element) - element) < 10**(-decimals):
            r_list.append(int(np.round(element)))
        else:
            r_list.append(np.round(element, decimals))

    return r_list


def standardize_vector(vector, prec=1e-5):
    """
    constructs a standard vector that indicates a direction (used in caches)

    :param vector: a vector
    :param prec: precision
    :return: standard vector
    """

    vector = np.array(vector) / np.linalg.norm(vector)  # 0.9999 # np.round(np.linalg.norm(axis), decimals=3)
    if abs(np.linalg.norm(vector)) < prec:
        raise Exception('Axis is too small, cannot normalize')

    return vector.tolist()


def rotate_basis_set(basis_set, angle, axis):
    """
    rotate a list of BasisFunction object

    :param basis_set: list of BasisFunction objects
    :param angle: angle to rotate
    :param axis: rotation axis
    :return: rotated basis set
    """
    new_basis_set = deepcopy(basis_set)
    for bf in new_basis_set:
        bf.apply_rotation(angle, axis)
    return new_basis_set


def translate_basis_set(basis_set, translation):
    """
    translate a list of BasisFunction object

    :param basis_set: list of BasisFunction objects
    :param translation: displacement vector
    :return: translated basis set
    """
    new_basis_set = deepcopy(basis_set)
    for bf in new_basis_set:
        bf.apply_translation(translation)
    return new_basis_set


def get_self_similarity(basis_set, density_matrix):
    """
    compute the self similarity of a density matrix expressed in the basis_set basis

    :param basis_set: list of BasisFunction objects
    :param density_matrix: density matrix
    :return: the self similarity
    """
    from sympy.utilities.iterables import multiset_permutations
    n = len(basis_set)
    s = np.zeros((n, n, n, n))

    for i in range(n):
        for j in range(i+1):
            for k in range(j+1):
                for l in range(k+1):
                    integral = (basis_set[i] * basis_set[j] * basis_set[k] * basis_set[l]).integrate
                    for perm in multiset_permutations([i, j, k, l]):
                        dens_prod = density_matrix[perm[0], perm[1]] * density_matrix[perm[2], perm[3]]
                        s[perm[0], perm[1], perm[2], perm[3]] = integral * dens_prod

    return np.sum(s)


def build_density(basis_set, density_matrix):
    """
    helper function to build a density function from a basis and a coefficients matrix (density matrix)

    :param basis_set: list of BasisFunction objects
    :param density_matrix: density matrix
    :return: the density matrix (BasisFunction object)
    """
    density_matrix = np.array(density_matrix)
    density = BasisFunction([], [])
    for i, basis1 in enumerate(basis_set):
        for j, basis2 in enumerate(basis_set):
            density += basis1*basis2 * density_matrix[i, j]

    return density


def build_orbital(basis_set, mo_coefficients):
    """
    helper function to build an orbital from a basis and coefficients

    :param basis_set: list of BasisFunction objects
    :param mo_coefficients: Molecular Orbital coefficients
    :return: the orbitals (BasisFunction object)
    """
    orbital = BasisFunction([], [])
    for mo_coeff, basis in zip(mo_coefficients, basis_set):
        orbital += mo_coeff * basis

    return orbital


def get_basis_set(coordinates, basis, use_angstrom=True):
    """
    get list of basis functions (basis_set) from pyQchem formatted basis dictionary

    :param coordinates: list/array of atom positions
    :param basis: pyQchem basis dictionary
    :param use_angstrom: if True, coordinates are provided in angstrom else Bohr
    :return: list of BasisFunction objects
    """
    if use_angstrom:
        coordinates = np.array(coordinates)*1.8897259886  # convert from Angstrom to bohr

    basis_list = []
    for iatom, atom in enumerate(basis['atoms']):
        for shell in atom['shells']:
            if shell['shell_type'] == 's':
                primitives = []
                for exponent in shell['p_exponents']:
                    primitives.append(PrimitiveGaussian(alpha=exponent))

                basis_list.append(BasisFunction(primitives, shell['con_coefficients'],
                                                center=coordinates[iatom],
                                                label='{}:S'.format(atom['symbol'])))

            elif shell['shell_type'] == 'sp':
                primitives = []
                for exponent in shell['p_exponents']:
                    primitives.append(PrimitiveGaussian(alpha=exponent))

                basis_list.append(BasisFunction(primitives, shell['con_coefficients'],
                                                center=coordinates[iatom],
                                                label='{}:S'.format(atom['symbol'])))

                for l_set in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))

                    basis_list.append(BasisFunction(primitives, shell['p_con_coefficients'],
                                                    center=coordinates[iatom],
                                                    label='{}:P_{}'.format(atom['symbol'], l_set)))

            elif shell['shell_type'] == 'p':
                for l_set in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))

                    basis_list.append(BasisFunction(primitives, shell['con_coefficients'],
                                                    center=coordinates[iatom],
                                                    label='{}:P_{}'.format(atom['symbol'], l_set)))

            elif shell['shell_type'] == 'd':
                for l_set in [[2, 0, 0], [0, 2, 0], [0, 0, 2], [1, 1, 0], [1, 0, 1], [0, 1, 1]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))
                    basis_list.append(BasisFunction(primitives, shell['con_coefficients'],
                                                    center=coordinates[iatom],
                                                    label='{}:D_{}'.format(atom['symbol'], l_set)))

            elif shell['shell_type'] == 'd_':
                basis_list_temp = []
                for l_set in [[2, 0, 0], [0, 2, 0], [0, 0, 2], [1, 1, 0], [1, 0, 1], [0, 1, 1]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))
                    basis_list_temp.append(BasisFunction(primitives, shell['con_coefficients'],
                                                         center=coordinates[iatom],
                                                         label='{}:D_{}'.format(atom['symbol'], l_set)))

                # d0 : 2*z2-x2-y2
                d0 = 2 * basis_list_temp[2] - basis_list_temp[0] - basis_list_temp[1]
                norm = 1/np.sqrt((d0*d0).integrate)
                basis_list.append(d0 * norm)

                # d1 : xz
                d1 = basis_list_temp[4]
                basis_list.append(d1)

                # d2 : yz
                d2 = basis_list_temp[5]
                basis_list.append(d2)

                # d3 : x2-y2
                d3 = basis_list_temp[0] - basis_list_temp[1]
                norm = 1/np.sqrt((d3*d3).integrate)
                basis_list.append(d3 * norm)

                # d4 : xy
                d4 = basis_list_temp[3]
                basis_list.append(d4)

            elif shell['shell_type'] == 'f':
                for l_set in [[3, 0, 0], [2, 1, 0], [2, 0, 1], [1, 2, 0], [1, 1, 1],
                              [1, 0, 2], [0, 3, 0], [0, 2, 1], [0, 1, 2], [0, 0, 3]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))
                    basis_list.append(BasisFunction(primitives, shell['con_coefficients'],
                                                    center=coordinates[iatom],
                                                    label='{}:F_{}'.format(atom['symbol'], l_set)))

            elif shell['shell_type'] == 'f_':
                basis_list_temp = []
                              #  xxx 0     xxy 1      xxz 2      xyy 3      xyz 4
                for l_set in [[3, 0, 0], [2, 1, 0], [2, 0, 1], [1, 2, 0], [1, 1, 1],
                              #  xzz 5     yyy 6      yyz 7      yzz 8      zzz 9
                              [1, 0, 2], [0, 3, 0], [0, 2, 1], [0, 1, 2], [0, 0, 3]]:
                    primitives = []
                    for exponent in shell['p_exponents']:
                        primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))
                    basis_list_temp.append(BasisFunction(primitives, shell['con_coefficients'],
                                                         center=coordinates[iatom],
                                                         label='{}:F_{}'.format(atom['symbol'], l_set)))

                # f0 : zzz - 3.0/2.0 * (xxz + yyz)
                f0 = basis_list_temp[9] - 3/2*(basis_list_temp[2] + basis_list_temp[7])
                norm = 1/np.sqrt((f0*f0).integrate)
                basis_list.append(f0 * norm)

                # f1 : 6.0 * xzz - 3.0/2.0 * (xxx + xyy)
                f1 = 6 * basis_list_temp[5] - 3/2 * (basis_list_temp[0] + basis_list_temp[3])
                norm = 1/np.sqrt((f1*f1).integrate)
                basis_list.append(f1 * norm)

                # f2 : 6.0 * yzz - 3.0/2.0 * (xxy + yyy)
                f2 = 6 * basis_list_temp[8] - 3/2 * (basis_list_temp[1] + basis_list_temp[6])
                norm = 1/np.sqrt((f2*f2).integrate)
                basis_list.append(f2 * norm)

                # f3 : 15.0 * (xxz - yyz))
                f3 = 15 * (basis_list_temp[2] - basis_list_temp[7])
                norm = 1/np.sqrt((f3*f3).integrate)
                basis_list.append(f3 * norm)

                # f4 : 30.0 * xyz
                f4 = 30 * basis_list_temp[4]
                norm = 1/np.sqrt((f4*f4).integrate)
                basis_list.append(f4 * norm)

                # f5 : 15.0 * xxx - 45.0 * xyy
                f5 = 15 * basis_list_temp[0] - 45 * basis_list_temp[3]
                norm = 1/np.sqrt((f5*f5).integrate)
                basis_list.append(f5 * norm)

                # f6 : 45.0 * xxy - 15.0 * yyy
                f6 = 45 * basis_list_temp[1] - 15 * basis_list_temp[6]
                norm = 1/np.sqrt((f6*f6).integrate)
                basis_list.append(f6 * norm)

            else:
                raise Exception('Not implemented shell type:{}'.format(shell['shell_type']))

    return basis_list


def get_basis_set_pyscf(pyscf_mol):
    """
    get list of basis functions (basis_set) from pyscf molecule object

    :param pyscf_mol: pyscf molecule object
    :return: list of basis functions
    """

    bf_names = ['s', 'p', 'd', 'f', 'g', 'h', 'i']

    basis_list = []
    for bas_id in range(pyscf_mol.nbas):

        if pyscf_mol.bas_angular(bas_id) == 0:  # s
            primitives = []
            for prim_id, exponent in enumerate(pyscf_mol.bas_exp(bas_id)):
                primitives.append(PrimitiveGaussian(alpha=exponent))

            basis_list.append(BasisFunction(primitives, pyscf_mol.bas_ctr_coeff(bas_id),
                                                 center=pyscf_mol.bas_coord(bas_id),
                                                 label='{}:{}'.format(pyscf_mol.atom_symbol(pyscf_mol.bas_atom(bas_id)),
                                                                      bf_names[pyscf_mol.bas_angular(bas_id)]))
                                   )

        elif pyscf_mol.bas_angular(bas_id) == 1:  # p
            for l_set in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
                primitives = []
                for prim_id, exponent in enumerate(pyscf_mol.bas_exp(bas_id)):
                    primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))


                basis_list.append(BasisFunction(primitives, pyscf_mol.bas_ctr_coeff(bas_id),
                                                     center=pyscf_mol.bas_coord(bas_id),
                                                     label='{}:{}'.format(pyscf_mol.atom_symbol(pyscf_mol.bas_atom(bas_id)),
                                                                          bf_names[pyscf_mol.bas_angular(bas_id)]))
                                       )

        elif pyscf_mol.bas_angular(bas_id) == 2:  # d

            basis_list_temp = []
            for l_set in [[2, 0, 0], [0, 2, 0], [0, 0, 2], [1, 1, 0], [1, 0, 1], [0, 1, 1]]:
                primitives = []
                for prim_id, exponent in enumerate(pyscf_mol.bas_exp(bas_id)):
                    primitives.append(PrimitiveGaussian(alpha=exponent, l=l_set))

                basis_list_temp.append(BasisFunction(primitives, pyscf_mol.bas_ctr_coeff(bas_id),
                                                     center=pyscf_mol.bas_coord(bas_id),
                                                     label='{}:{}'.format(pyscf_mol.atom_symbol(pyscf_mol.bas_atom(bas_id)),
                                                                          bf_names[pyscf_mol.bas_angular(bas_id)])))

            # d1 : xy
            d1 = basis_list_temp[3]
            basis_list.append(d1)

            # d2 : yz
            d2 = basis_list_temp[5]
            basis_list.append(d2)

            # d3 : 2*z2-x2-y2 (z^2?)
            d3 = 2 * basis_list_temp[2] - basis_list_temp[0] - basis_list_temp[1]
            norm = 1 / np.sqrt((d3 * d3).integrate)
            basis_list.append(d3 * norm)

            # d4 : xz
            d4 = basis_list_temp[4]
            basis_list.append(d4)

            # d5 : x2-y2
            d5 = basis_list_temp[0] - basis_list_temp[1]
            norm = 1/np.sqrt((d5*d5).integrate)
            basis_list.append(d5 * norm)

        else:
            raise Exception('Not implemented shell type with angular momentum:{}'.format(pyscf_mol.bas_angular(bas_id)))

    return basis_list


def get_principal_axis_angles(cm_vectors, masses=None):
    """
    Compute the principal axis and moments of inertia

    :param cm_vectors: list of distance vectors of atoms respect center of mass
    :param masses: list of atomic masses (if None then set all to 1)
    :return: euler angles that alings the structure along the principal axis of inertia
    """

    if masses is None:
        masses = np.ones(len(cm_vectors))

    # Build inertia tensor
    inertia_tensor = np.zeros((3, 3))
    for m, c in zip(masses, cm_vectors):
        inertia_tensor += m * (np.identity(3) * np.dot(c, c) - np.outer(c, c))

    # Compute eigenvalues and eigenvectors of inertia tensor
    e_values, e_vectors = np.linalg.eigh(inertia_tensor)  # be careful eigenvectors in columns!
    axis_of_inertia = e_vectors.T

    # return align euler angles
    return R.from_matrix(axis_of_inertia).as_euler('zyx', degrees=True)


# grid generators
def uniform_euler_scan(val, step):
    """
    generates a uniform scan of Euler angles from -val to +val in 3d

    :param val: scan bounds
    :param step: step of the scan
    :return: list of 3 Euler angles
    """
    ranges = np.arange(-val, val + step, step)
    for angles in itertools.product(ranges, ranges, ranges):
        yield angles


def skewed_euler_scan(val, step):
    """
    generates a skewed scan of Euler angles from -val to +val in 3d.
    This tries to approximate a uniform spherical grid around a point

    :param val: scan bounds
    :param step: step of the scan
    :return: list of 3 Euler angles
    """
    ranges = np.arange(-val, val + step, step)
    for angles_2d in itertools.product(ranges, ranges):
        step_final = step / np.max([np.sin(np.deg2rad(angles_2d[1])), step / 180])
        for angle in np.arange(-val, val + step_final, step_final):
            yield np.array(list(angles_2d) + [angle], dtype=float)


def collapse_limit(coordinates, tolerance=1e-5):
    if np.average(np.linalg.norm(coordinates, axis=1)) < tolerance:
        return True
    else:
        return False


__author__ = 'Abel Carreras'
__version__ = '1.2.0'

from posym.tools import list_round, get_principal_axis_angles
from posym.pointgroup import PointGroup
from posym.basis import BasisFunction
from posym.config import Configuration
from posym.tools import uniform_euler_scan, collapse_limit
from posym.permutation import generate_permutation_set
from posym.permutation.hungarian import get_permutation_hungarian
from scipy.spatial.transform import Rotation as R
from scipy.optimize import minimize
import numpy as np
import pandas as pd
import warnings


cache_orientation = {}
def get_hash(coordinates, symbols, group):
    return hash((np.array2string(coordinates, precision=2), tuple(symbols), group))


class SymmetryObject:
    """
    Main symmetry object that abstracts an element in the g-module space

    """
    def __init__(self, group, rep, normalize=False):

        self._pg = PointGroup(group)
        self._group = group.lower()

        if isinstance(rep, str):
            if rep not in self._pg.ir_labels:
                raise Exception('Representation do not match with group\nAvailable: {}'.format(self._pg.ir_labels))
            self._op_representation = self._pg.ir_table[rep]
            #if normalize:
            #    self._op_representation /= self._pg.ir_table[rep]['E']

        elif isinstance(rep, pd.Series):
            if np.all(self._pg.ir_table.sort_index().index == rep.sort_index().index):
                self._op_representation = rep.reindex(self._pg.ir_table.index)
            else:
                raise Exception('Representation not in group')

        if normalize:
            op_rep = np.dot(self._pg.trans_matrix_norm, np.dot(self._pg.trans_matrix_inv, self._op_representation.values))
            self._op_representation = pd.Series(op_rep, index=self._pg.op_labels)

    def get_reduced_op_representation(self):
        red_values = []
        for value in self._op_representation.values:
            red_values.append(np.average(value))
        return pd.Series(red_values, index=self._op_representation.index)

    def get_op_representation(self):
        return self._op_representation

    def get_ir_representation(self):
        ir_rep = np.dot(self._pg.trans_matrix_inv, self.get_reduced_op_representation().values)
        return pd.Series(ir_rep, index=self._pg.ir_labels)

    def get_point_group(self):
        return self._pg

    def __str__(self):

        ir_rep = self.get_ir_representation().values
        ir_rep = list_round(ir_rep)

        ir_labels = self.get_ir_representation().index

        str = ''
        for i, r in enumerate(ir_rep):
            if np.add.reduce(np.square(ir_rep[:i])) > 0 and r > 0:
                    str += ' + '
            elif r < 0:
                    str += ' - '
            if np.abs(r - 1) < 2e-2:
                str += ir_labels[i]
            elif np.abs(r) > 0:
                str += '{} {}'.format(abs(r), ir_labels[i])

        return str

    def __repr__(self):
        ir_rep = self.get_ir_representation().values
        ir_rep = list_round(ir_rep)

        ir_labels = self.get_ir_representation().index

        txt = ''
        for i, r in enumerate(ir_rep):
            # print('>> ', np.add.reduce(ir_rep[:i]**2), 'r:', r)
            if np.add.reduce(np.square(ir_rep[:i])) > 0 and r > 0 and len(ir_rep[:i]) > 0:
                txt += '+'
            if r == 1:
                txt += ir_labels[i]
            elif r > 0:
                txt += '{}{}'.format(r, ir_labels[i])
        return txt

    def __add__(self, other):

        if self._group == other._group:
            return SymmetryObject(self._group,
                                  self._op_representation + other._op_representation)

        raise Exception('Incompatible point groups')

    def __rmul__(self, other):
        return self.__mul__(other)

    def __mul__(self, other):

        if isinstance(other, (float, int)):
            return SymmetryObject(self._group,
                                  self._op_representation * other)

        elif isinstance(other, SymmetryObject):
            mul_rep = self._op_representation * other._op_representation

            return SymmetryObject(self._group, mul_rep)
        else:
            raise Exception('Symmetry operation not possible')


class SymmetryMolecule(SymmetryObject):
    """
    Symmetry of molecular geometry
    """
    def __init__(self, group, coordinates, symbols, total_state=None, orientation_angles=None, center=None, permutation_set=None):
        """

        :param group: symmetry point group
        :param coordinates: molecular coordinates
        :param symbols: atomic symbols
        :param total_state: symmetry operations overlaps (SOEV's) as a panda Series object
        :param orientation_angles: orientation angles
        :param center: center of symmetry group [x, y, z]
        """

        self._setup_structure(coordinates, symbols, group, center, orientation_angles, permutation_set=permutation_set)

        if total_state is None:
            # m = self.measure_pos
            self._generate_permutation_set(self._angles)
            rotmol = R.from_euler('zyx', self._angles, degrees=True)
            self._operator_measures = []
            for operation in self._pg.operations:
                operator_measures = []
                for op in self._pg.get_sub_operations(operation.label):

                    # check if all atoms are collapsed in a point
                    if collapse_limit(self._coordinates):
                        operator_measures.append(1)
                        continue

                    overlap = op.get_measure_pos(self._coordinates, orientation=rotmol)
                    operator_measures.append(overlap)

                # print('operator measures', operator_measures)
                self._operator_measures.append(np.array(operator_measures))

            total_state = pd.Series(self._operator_measures, index=self._pg.op_labels)

        if not self.check_permutation_coherence and not collapse_limit(self.symmetrized_coordinates):
            warnings.warn('Incoherence found in symmetrized structure. Symmetry measure may be incorrect')

        super().__init__(group, total_state)

    def _setup_structure(self, coordinates, symbols, group, center, orientation_angles, permutation_set=None):

        conf = Configuration()

        self._coordinates = np.array(coordinates)
        self._symbols = symbols
        self._pg = PointGroup(group)
        self._permutation_set = {}

        if '_center' not in self.__dir__():
            self._center = center
            if self._center is None:
                self._center = np.average(self._coordinates, axis=0)

            self._coordinates = np.array([c - self._center for c in self._coordinates])

        if orientation_angles is None:
            self._angles = self.get_orientation(fast_optimization=conf.fast_optimization, scan_step=conf.scan_steps)
        else:
            self._angles = orientation_angles

        # manual permutation
        if permutation_set is not None:
            self._permutation_set[tuple(orientation_angles)] = {gen: perm for gen, perm in
                                                                zip(self._pg.generators,permutation_set)}

        self._generate_permutation_set(self._angles)

    def get_orientation(self, fast_optimization=True, scan_step=20, guess_angles=None):
        """
        get orientation angles for optimum orientation.
        Use full=False to orient perfect symmetric molecules. Use full=True to orient quasi symmetric molecules

        :param fast_optimization: if True use only a subset of symmetry elements (for exact symmetry objets)
        :param scan_step: step angle (deg) use for the preliminary scan
        :return:
        """

        hash_num = get_hash(self._coordinates, self._symbols, self._pg.label)
        if hash_num in cache_orientation:
            return cache_orientation[hash_num]

        # optimization functions
        def optimization_function_simple(angles):
            """
            This function uses only one operation of each type (described in the IR table).
            This approach works well when the molecule has a symmetry close to the group
            """
            self._generate_permutation_set(angles)
            rotmol = R.from_euler('zyx', angles, degrees=True)

            coor_measures = []
            for operation in self._pg.operations:
                coor_m = operation.get_measure_pos(self._coordinates, orientation=rotmol, normalized=False)
                coor_measures.append(coor_m)

            # get most symmetric IR value
            return -np.dot(coor_measures, self._pg.trans_matrix_inv[0])

        def optimization_function_full(angles):
            """
            This function uses all operations of the group and averages the overlap of equivalent operations
            """

            self._generate_permutation_set(angles)
            rotmol = R.from_euler('zyx', angles, degrees=True)

            operator_measures = []
            for operation in self._pg.operations:
                sub_operator_measures = []
                for op in self._pg.get_sub_operations(operation.label):
                    overlap = op.get_measure_pos(self._coordinates, orientation=rotmol)
                    sub_operator_measures.append(overlap)
                operator_measures.append(np.average(sub_operator_measures))

            # get most symmetric IR value
            return -np.dot(operator_measures, self._pg.trans_matrix_inv[0])

        # define if use simple function (faster) or full (slower)
        optimization_function = optimization_function_simple if fast_optimization else optimization_function_full

        # define better orientation for preliminar scan
        pai_angles = get_principal_axis_angles(self._coordinates)

        # preliminary scan
        if guess_angles is None:
            guess_angles = ref_value = None
            for angles in uniform_euler_scan(90, scan_step):
                angles -= pai_angles
                value = optimization_function(angles)
                if ref_value is None or value < ref_value:
                    ref_value = value
                    guess_angles = angles

        result = minimize(optimization_function, guess_angles, method='CG')

        cache_orientation[hash_num] = result.x
        return cache_orientation[hash_num]

    def get_oriented_operations(self):
        import copy
        rotmol = R.from_euler('zyx', self.orientation_angles, degrees=True)

        operations_list = []
        for operation in self._pg.operations:
            for sub_operation in self._pg.get_sub_operations(operation.label):
                sub_operation = copy.deepcopy(sub_operation)
                sub_operation.apply_rotation(rotmol)
                operations_list.append(sub_operation)

        return operations_list

    def print_operations_info(self):
        from posym.permutation import Permutation

        print('\nOperations list (molecule orientation)'
              '\n--------------------------------------')
        self._generate_permutation_set(self._angles)
        for op in self.get_oriented_operations():
            # print('Operation object: ', op)
            print('label:', op.label)
            try:
                print('Order:', op.order)
                print('Exponent:', op.exp)
            except AttributeError:
                pass
            try:
                print('Axis:', op.axis)
            except AttributeError:
                pass

            print('permutation:', op.permutation)
            permutation = Permutation(op.permutation)
            print('orbits: ', permutation.get_orbits())

            print('Matrix representation:')
            print(np.round(op.matrix_representation, decimals=6))
            print()

    def _generate_permutation_set(self, angles, force_reset=False):

        rotmol = R.from_euler('zyx', angles, degrees=True)
        dict_key = tuple(angles)

        if dict_key not in self._permutation_set or force_reset:

            if collapse_limit(self._coordinates):
                self._permutation_set[dict_key] = next(generate_permutation_set(self._pg.generators, self._symbols))
                return

            # Hungarian algorithm (approximated)
            if Configuration().algorithm == 'hungarian':
                permutation_set = {}
                for gen in self._pg.generators:
                    rot_coor = rotmol.inv().apply(self._coordinates)
                    permutation_set[gen] = get_permutation_hungarian(gen.matrix_representation, rot_coor, self._symbols)
                self._permutation_set[dict_key] = permutation_set

            # Brute force algorithm (exact)
            elif Configuration().algorithm == 'exact':
                ir_rep_diff_max = -100

                class NotValidPermutation(Exception): pass

                for permutation_set in generate_permutation_set(self._pg.generators, self._symbols):

                    try:
                        operator_measures = []
                        for operation in self._pg.operations:
                            sub_operator_measures = []
                            for op in self._pg.get_sub_operations(operation.label):
                                if op.set_permutation_set(permutation_set, self._symbols) is None:
                                    raise NotValidPermutation

                                overlap = op.get_measure_pos(self._coordinates, orientation=rotmol)
                                sub_operator_measures.append(overlap)

                            operator_measures.append(np.average(sub_operator_measures))

                        ir_rep_diff = np.dot(operator_measures, self._pg.trans_matrix_inv[0])

                        if ir_rep_diff_max < ir_rep_diff:
                            ir_rep_diff_max = ir_rep_diff
                            self._permutation_set[dict_key] = permutation_set

                    except NotValidPermutation:
                        continue
            else:
                raise Exception('Permutation algorithm not recognized ')

        for operation in self._pg.operations:
            operation.set_permutation_set(self._permutation_set[dict_key], self._symbols, ignore_compatibility=True)
            for op in self._pg.get_sub_operations(operation.label):
                op.set_permutation_set(self._permutation_set[dict_key], self._symbols, ignore_compatibility=True)

    @property
    def measure(self):
        norm = self.get_reduced_op_representation().values[0]
        return 100*(1-np.array(self.get_ir_representation().values[0])/norm)

    @property
    def measure_pos(self):

        if collapse_limit(self._coordinates):
            return 0.0

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        self._generate_permutation_set(self._angles)

        operator_measures = []
        for operation in self._pg.operations:
            sub_operator_measures = []
            for op in self._pg.get_sub_operations(operation.label):
                overlap = op.get_measure_pos(self._coordinates, orientation=rotmol)
                sub_operator_measures.append(overlap)

            operator_measures.append(np.average(sub_operator_measures))

        # get most symmetric IR value
        ir_rep_diff = np.dot(operator_measures, self._pg.trans_matrix_inv[0])

        # return csm
        return 100 * (1 - ir_rep_diff)

    @property
    def check_permutation_coherence(self, tolerance=0.95):
        """
        check coherence of the permutation by checking the symmetry measure of
        the symmetrized structure.  This function only checks the symmetrized structure.
        The measure may still be correct.

        :param tolerance: tolenrece value for the final overlap measure of the structure
        :return: True or False
        """

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        operator_measures = []
        for operation in self._pg.operations:
            sub_operator_measures = []
            for op in self._pg.get_sub_operations(operation.label):
                if collapse_limit(self._coordinates):
                    sub_operator_measures.append(1)
                    continue

                overlap = op.get_measure_pos(self.symmetrized_coordinates, orientation=rotmol)
                sub_operator_measures.append(overlap)

            operator_measures.append(np.average(sub_operator_measures))

        # get most symmetric IR value
        ir_rep_diff = np.dot(operator_measures, self._pg.trans_matrix_inv[0])

        return ir_rep_diff > tolerance

    @property
    def opt_coordinates(self):
        rotmol = R.from_euler('zyx', self._angles, degrees=True)
        return rotmol.apply(self._coordinates)

    @property
    def symmetrized_coordinates(self):

        self._generate_permutation_set(self._angles)
        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        structure_list = []
        for op in self._pg.all_operations:
            structure = op.get_operated_coordinates(self._coordinates, orientation=rotmol)
            structure_list.append(structure)

        return np.average(structure_list, axis=0)

    @property
    def orientation_angles(self):
        return self._angles

    @property
    def center(self):
        return self._center


class SymmetryNormalModes(SymmetryMolecule):
    """
    get symmetry of the normal modes.
    """
    def __init__(self, group, coordinates, modes, symbols, orientation_angles=None, center=None):
        """

        :param group: symmetry point group
        :param coordinates: atomic coordinates
        :param modes: list of normal modes separated by atoms [[[x1, y1, z1], [x2, y2, z2], [x3, y3, z3]], ...]
        :param symbols: atomic symbols
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        :param center: center of symmetry group [x, y, z]
        """

        self._setup_structure(coordinates, symbols, group, center, orientation_angles)

        self._modes = modes

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        self._mode_measures = []
        for operation in self._pg.operations:
            mode_measures = []
            for op in self._pg.get_sub_operations(operation.label):
                mode_m = op.get_measure_modes(self._modes, orientation=rotmol)
                mode_measures.append(mode_m)

            mode_measures = np.array(mode_measures)
            self._mode_measures.append(mode_measures)

        mode_measures_total = []
        for op in self._mode_measures:
            op_list = []
            for m in op:
                op_list.append(sum(m))
            mode_measures_total.append(op_list)

        # reshape mode measures
        reshaped_modes_measures = []
        for m in range(len(self._mode_measures[0].T)):
            reshaped_modes_measures.append([k[:, m] for k in self._mode_measures])

        self._mode_measures = reshaped_modes_measures

        total_state = pd.Series(mode_measures_total, index=self._pg.op_labels)

        super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])

    def get_state_mode(self, n):
        return SymmetryObject(group=self._group, rep=pd.Series(self._mode_measures[n],
                                                               index=self._pg.op_labels))

    def get_number_of_modes(self):
        return len(self._mode_measures)


class SymmetryAtomDisplacements(SymmetryMolecule):
    """
    get symmetry of atom displacements . This is equivalent to the sum of the symmetries of all
    normal modes (gamma 3N).
    """
    def __init__(self, group, coordinates, symbols, orientation_angles=None):
        """

        :param group: symmetry group
        :param coordinates: atomic coordinates
        :param symbols: atomic symbols
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        """

        self._setup_structure(coordinates, symbols, group, None, orientation_angles)

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        trans_rots = []
        for label in self._pg.ir_table.rotations + self._pg.ir_table.translations:
            trans_rots.append(self._pg.ir_table[label].values)

        trans_rots = np.sum(trans_rots, axis=0)

        self._mode_measures = []
        for operation in self._pg.operations:
            mode_measures = []
            for op in self._pg.get_sub_operations(operation.label):
                measure_xyz = op.get_measure_xyz(orientation=rotmol)
                measure_atom = op.get_measure_atom()
                mode_measures.append(measure_xyz * measure_atom)

            mode_measures = np.array(mode_measures)
            self._mode_measures.append(mode_measures)

        self._mode_measures = np.array(self._mode_measures, dtype=object).flatten() - trans_rots
        total_state = pd.Series(self._mode_measures, index=self._pg.op_labels)
        super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])


class SymmetryAdaptedCoordinates(SymmetryMolecule):
    """
    get symmetry of atom displacements . This is equivalent to the sum of the symmetries of all
    normal modes (gamma 3N).
    """
    def __init__(self, group, coordinates, symbols, orientation_angles=None):
        """

        :param group: symmetry group
        :param coordinates: atomic coordinates
        :param symbols: atomic symbols
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        """

        self._setup_structure(coordinates, symbols, group, None, orientation_angles)

        if not self._pg.is_abelian:
            raise Exception('Not implemented for non-Abelian groups')

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        self._modes = []
        self._op_overlaps = []
        for traces_vector in self._pg.trans_matrix.T:
            projection = np.zeros((3 * len(coordinates), len(coordinates), 3))
            for operation, trace in zip(self._pg.operations, traces_vector):
                projection += trace * operation.get_displacements_projection(orientation=rotmol)

            def linear_indepedent(vect):
                for m in self._modes:
                    dot = np.abs(np.abs(np.dot(m.flatten(), vect.flatten())) - 1)
                    if dot < 1e-5:
                        return False
                return True

            for m in projection:
                norm = np.linalg.norm(m.flatten())
                if abs(norm) > 1e-5 and linear_indepedent(m/norm):
                    self._modes.append(m/norm)
                    self._op_overlaps.append(traces_vector)

        total_state = pd.Series(np.sum(self._op_overlaps, axis=0), index=self._pg.op_labels)
        super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])

    def get_symmetry_adapted_coordinates(self):
        return np.array(self._modes).tolist()

    def get_state_mode(self, n):
        return SymmetryObject(group=self._group, rep=pd.Series(self._op_overlaps[n],
                                                               index=self._pg.op_labels))

    def get_number_of_modes(self):
        return len(self._op_overlaps)


class SymmetryGaussianLinear(SymmetryMolecule):
    """
    get symmetry from a function defined in the basis of Gaussian functions (BasisFunction object)

    """
    def __init__(self, group, function, orientation_angles=None, center=None):
        """

        :param group: symmetry group
        :param function: the function (BasisFunction object)
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        :param center: center of symmetry group [x, y, z]
        """

        symbols, coordinates = function.get_environment_centers()

        self._setup_structure(coordinates, symbols, group, center, orientation_angles)

        self._function = function.copy()
        self._function.apply_translation(-np.array(self._center))

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        self._self_similarity = (self._function * self._function).integrate

        self._operator_measures = []
        for operation in self._pg.operations:
            operator_measures = []
            for op in self._pg.get_sub_operations(operation.label):
                overlap = op.get_overlap_func(self._function, self._function, orientation=rotmol)
                operator_measures.append(overlap/self._self_similarity)

            self._operator_measures.append(np.array(operator_measures))

        total_state = pd.Series(self._operator_measures, index=self._pg.op_labels)
        super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])

    @property
    def self_similarity(self):
        return self._self_similarity


class SymmetrySingleDeterminant(SymmetryMolecule):
    """
    get symmetry from single determinat wave function
    """
    def __init__(self, group, alpha_orbitals, beta_orbitals, orientation_angles=None, center=None):
        """

        :param group: symmetry group
        :param alpha_orbitals: list of alpha orbitals (BasisFunction objects)
        :param beta_orbitals: list of beta orbitals (BasisFunction objects)
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        :param center: center of symmetry group [x, y, z]
        """

        # generate copy to not modify original orbitals
        alpha_orbitals = [f.copy() for f in alpha_orbitals]
        beta_orbitals = [f.copy() for f in beta_orbitals]

        function = BasisFunction([], [])
        for f in alpha_orbitals:
            function = function + f
        for f in beta_orbitals:
            function = function + f

        symbols, coordinates = function.get_environment_centers()

        self._setup_structure(coordinates, symbols, group, center, orientation_angles)

        data_set = set(list(alpha_orbitals) + list(beta_orbitals))
        for f in data_set:
            f.apply_translation(-np.array(self._center))

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        def get_overlaps(orbitals):

            operator_overlaps_total = []
            for i, a_orb in enumerate(orbitals):
                overlap_row = []
                for j, b_orb in enumerate(orbitals):
                    self_similarity = (a_orb*a_orb).integrate * (b_orb*b_orb).integrate

                    overlaps_list = []
                    for operation in self._pg.operations:
                        operator_overlaps = []
                        for op in self._pg.get_sub_operations(operation.label):
                            overlap = op.get_overlap_func(a_orb, b_orb, orientation=rotmol)
                            operator_overlaps.append(overlap/self_similarity)

                        operator_overlaps = np.array(operator_overlaps)
                        # operator_overlaps = np.average(operator_overlaps, axis=0)
                        overlaps_list.append(operator_overlaps)

                    overlap_row.append(overlaps_list)
                operator_overlaps_total.append(overlap_row)

            operator_overlaps = []
            for k in range(self._pg.n_ir):
                multi_over = []
                n_degenerate = len(operator_overlaps_total[0][0][k])
                for m in range(n_degenerate):
                    matrix = np.zeros((len(orbitals), len(orbitals)))
                    for i in range(len(orbitals)):
                        for j in range(len(orbitals)):
                            matrix[i, j] = operator_overlaps_total[i][j][k][m]

                    multi_over.append(np.linalg.det(matrix))
                operator_overlaps.append(np.array(multi_over, dtype=float))

            return operator_overlaps

        if len(alpha_orbitals) > 0 and len(beta_orbitals) > 0:

            operator_overlaps_alpha = get_overlaps(alpha_orbitals)
            operator_overlaps_beta = get_overlaps(beta_orbitals)

            total_state = pd.Series(operator_overlaps_alpha, index=self._pg.op_labels) * \
                          pd.Series(operator_overlaps_beta, index=self._pg.op_labels)

            super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])

        elif len(alpha_orbitals) > 0:

            operator_overlaps_alpha = get_overlaps(alpha_orbitals)
            total_state = pd.Series(operator_overlaps_alpha, index=self._pg.op_labels)
            super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])

        elif len(beta_orbitals) > 0:

            operator_overlaps_beta = get_overlaps(beta_orbitals)
            total_state = pd.Series(operator_overlaps_beta, index=self._pg.op_labels)
            super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, [0, 0, 0])


class SymmetryMultiDeterminant(SymmetryMolecule):
    """
    Get symmetry from multi determinant wave function
    """
    def __init__(self, group, orbitals, configurations, orientation_angles=None, center=None):
        """

        :param group: symmetry group
        :param orbitals: list BasisFunction objects
        :param configurations: dictionary that contains the electronic configuration (see README for example)
        :param orientation_angles: list of 3 Euler angles [pitch, yaw, roll]
        :param center: center of symmetry group [x, y, z]
        """
        # generate copy to not modify original orbitals
        orbitals = [f.copy() for f in orbitals]

        function = BasisFunction([], [])
        for f in orbitals:
            function = function + f

        symbols, coordinates = function.get_environment_centers()
        self._setup_structure(coordinates, symbols, group, center, orientation_angles)

        for f in orbitals:
            f.apply_translation(-np.array(self._center))

        rotmol = R.from_euler('zyx', self._angles, degrees=True)

        def set_linear_combination(overlaps_matrix, configurations):

            def get_sub_matrix(occupations1, occupations2):
                cmat = overlaps_matrix.copy()
                n = len(occupations1)
                for i, (c1, c2) in enumerate(zip(occupations1[::-1], occupations2[::-1])):
                    if c1 == 0:
                        cmat = np.delete(cmat, n - i - 1, 0)
                    if c2 == 0:
                        cmat = np.delete(cmat, n - i - 1, 1)
                return cmat

            overlap = 0
            for i, conf1 in enumerate(configurations):
                for j, conf2 in enumerate(configurations):
                    new_matrix_alpha = get_sub_matrix(conf1['occupations']['alpha'], conf2['occupations']['alpha'])
                    overlap_alpha = np.linalg.det(new_matrix_alpha)

                    new_matrix_beta = get_sub_matrix(conf1['occupations']['beta'], conf2['occupations']['beta'])
                    overlap_beta = np.linalg.det(new_matrix_beta)

                    overlap += overlap_alpha * overlap_beta * conf1['amplitude']*conf2['amplitude']
            return overlap

        def get_overlaps(orbitals, configurations):

            operator_overlaps_total = []
            for i, a_orb in enumerate(orbitals):
                overlap_row = []
                for j, b_orb in enumerate(orbitals):
                    self_similarity = (a_orb*a_orb).integrate * (b_orb*b_orb).integrate

                    overlaps_list = []
                    for operation in self._pg.operations:
                        operator_overlaps = []
                        for op in self._pg.get_sub_operations(operation.label):
                            overlap = op.get_overlap_func(a_orb, b_orb, orientation=rotmol)
                            operator_overlaps.append(overlap/self_similarity)

                        operator_overlaps = np.array(operator_overlaps)
                        # operator_overlaps = np.average(operator_overlaps, axis=0)
                        overlaps_list.append(operator_overlaps)

                    overlap_row.append(overlaps_list)
                operator_overlaps_total.append(overlap_row)

            operator_overlaps = []
            for k in range(self._pg.n_ir):
                multi_over = []
                n_degenerate = len(operator_overlaps_total[0][0][k])
                for m in range(n_degenerate):
                    matrix = np.zeros((len(orbitals), len(orbitals)))
                    for i in range(len(orbitals)):
                        for j in range(len(orbitals)):
                            matrix[i, j] = operator_overlaps_total[i][j][k][m]

                    multi_over.append(set_linear_combination(matrix, configurations))
                operator_overlaps.append(np.array(multi_over, dtype=float))

            return operator_overlaps

        operator_overlaps = get_overlaps(orbitals, configurations)
        total_state = pd.Series(operator_overlaps, index=self._pg.op_labels)
        super().__init__(group, self._coordinates, self._symbols, total_state, self._angles, center=[0, 0, 0])


if __name__ == '__main__':

    from posym.algebra import norm

    coordinates = [[ 0.000000000+00,  0.000000000+00,  2.40297090e-01],
                   [-1.43261539e+00, -1.75444785e-16, -9.61188362e-01],
                   [ 1.43261539e+00,  1.75444785e-16, -9.61188362e-01]]

    symbols = ['O', 'H', 'H']

    sm = SymmetryAtomDisplacements('c2v', coordinates, symbols)
    print(sm.get_point_group())
    print(sm)
    mb = SymmetryMolecule('c2v', coordinates, symbols)
    print('Coor measure: ', mb, '(', norm(mb), ')')



import pandas as pd
import numpy as np
from posym.operations.identity import Identity
from posym.operations.rotation import Rotation
from posym.operations.reflection import Reflection
from posym.operations.inversion import Inversion
from posym.operations.irotation import ImproperRotation


def real_radical(m, n):
    return 2 * np.cos(2 * m * np.pi / n)


class CharTable(pd.DataFrame):
    """
    Subclass of DataFrame that contains the symmetry table data

    """
    def __init__(self, name, operations, ir, rotations, translations, multiplicities, generator_indices=()):

        labels = [o.label for o in operations]

        data = {}
        for key, val in ir.items():
            val.index = labels
            data[key] = val

        super().__init__(data)
        self.name = name

        if np.all([r in data for r in rotations]):
            self.attrs['rotations'] = rotations
        else:
            raise Exception('Rotations representations not in table')

        if np.all([r in data for r in translations]):
            self.attrs['translations'] = translations
        else:
            raise Exception('Translations representations not in table')

        if len(multiplicities) == len(self.index):
            self.attrs['multiplicities'] = multiplicities
        else:
            raise Exception('Multiplicities do not match')

        self.attrs['operations'] = operations

        # define generators
        generator_list = []
        for i in generator_indices:
            generator = operations[i]
            generator._gen_rep = [generator]  # set itself as generators representation
            generator_list.append(generator)

        self.attrs['generators'] = generator_list

    def __str__(self):

        table = self.copy()
        new_index = []
        for op_label, m in zip(table.index, table.attrs['multiplicities']):
            if m > 1:
                new_index.append('{}{}'.format(m, op_label))
            else:
                new_index.append('{}'.format(op_label))

        table.index = new_index
        formatted_data = {k: "{:,.4f}".format for k in table.T.columns}
        # formatted_data = {k: "{:}".format for k in table.T.columns}  # no format

        return table.T.to_string(formatters=formatted_data)

    def get_all_operations(self):

        if 'all_operations' in self.attrs:
            return self.attrs['all_operations']

        # generate all operations of the group and store its relation in connection
        def add_layer(opt_list, opi):
            for gen_op in self.generators:
                op_p = gen_op * opi
                if not op_p in opt_list:
                    opt_list.append(op_p)

        def generate_operations(opt_list):

            while True:
                list_ini = list(opt_list)
                for op in opt_list:
                    add_layer(opt_list, op)

                if len(list_ini) == len(opt_list):
                    break

        opt_list = [self.operations[0]]
        generate_operations(opt_list)

        """
        print('n_operations:', len(opt_list))

        print('\nOperators from generators')
        for op in opt_list:
            try:
                print('connect: ', op.label, op.axis, op.order, op.exp, op)
            except:
                try:
                    print('connect: ', op.label, op.axis)
                except:
                    print('connect: ', op.label)
        """

        # set label to operations according to classes
        operation_dict = {}
        for op_ref in self.operations:
            # print('ref: ', op_ref.label, '               ', op_ref)
            operation_dict[op_ref.label] = []
            for op in opt_list:
                if op.get_type() == op_ref.get_type():
                    for op_b in opt_list:
                        prod = op_b * op * op_b.inverse()
                        if (op_ref == prod) or (op_ref == prod.inverse()):
                            #print('   found!', op)
                            #print('           check:', op_b * op * op_b.inverse())
                            op._label = op_ref.label
                            operation_dict[op_ref.label].append(op)
                            break

        # add gen_rep to operations
        for op_ref in self.operations:
            for op in opt_list:
                if op_ref == op:
                    op_ref._gen_rep = op._gen_rep

        self.attrs['all_operations'] = operation_dict

        return self.attrs['all_operations']

    @property
    def rotations(self):
        return [ir for ir in self.attrs['rotations']]

    @property
    def translations(self):
        return [ir for ir in self.attrs['translations']]

    @property
    def multiplicities(self):
        return self.attrs['multiplicities']

    @property
    def operations(self):
        return self.attrs['operations']

    @property
    def generators(self):
        return self.attrs['generators']

    @property
    def ir_degeneracies(self):
        return self.T['E'].values


ir_table_list = [

    CharTable('Cs',
              [Identity(label='E'), Reflection(label='sh', axis=[0, 0, 1])],
              {"A'": pd.Series([+1, +1]),
               "A''": pd.Series([+1, -1])
               },
              rotations=["A'", "A'", "A''"],  # x, y, z
              translations=["A''", "A''", "A'"],  # Rx, Ry, Rz
              multiplicities=[1, 1],
              generator_indices=[1]),

    CharTable('Ci',
              [Identity(label='E'), Inversion(label='i')],
              {'Ag': pd.Series([+1, +1]),
               'Au': pd.Series([+1, -1])
               },
              rotations=['Au', 'Au', 'Au'],  # x, y, z
              translations=['Ag', 'Ag', 'Ag'],  # Rx, Ry, Rz
              multiplicities=[1, 1],
              generator_indices=[1]),

    CharTable('T',
              [Identity(label='E'), Rotation(label='C3', axis=[0, 0, 1], order=3),
               Rotation(label='C2', axis=[np.sqrt(2 / 9), 0, 1 / 3], order=2)],
              {'A': pd.Series([+1, +1, +1]),
               'E': pd.Series([+2, -1, +2]),
               'T': pd.Series([+3,  0, -1])
               },
              rotations=['T'],
              translations=['T'],
              multiplicities=[1, 8, 3],
              generator_indices=[1, 2]),

    CharTable('Td',
              [Identity(label='E'), Rotation(label='C3', axis=[0, 0, 1], order=3),
               Rotation(label='C2', axis=[np.sqrt(2/9), 0, 1/3], order=2),
               ImproperRotation(label='S4', axis=[np.sqrt(2/9), 0, 1/3], order=4),
               Reflection(label='sd', axis=[0, 1, 0])],
              {'A1': pd.Series([+1, +1, +1, +1, +1]),
               'A2': pd.Series([+1, +1, +1, -1, -1]),
               'E' : pd.Series([+2, -1, +2,  0,  0]),
               'T1': pd.Series([+3,  0, -1, +1, -1]),
               'T2': pd.Series([+3,  0, -1, -1, +1])
               },
              rotations=['T1'],
              translations=['T2'],
              multiplicities=[1, 8, 3, 6, 6],
              generator_indices=[1, 3]),

    CharTable('Th',
              [Identity(label='E'), Rotation(label='C3', axis=[0, 0, 1], order=3),
               Rotation(label='C2', axis=[np.sqrt(2/9), 0, 1/3], order=2),
               Inversion(label='i'),
               ImproperRotation(label='S6', axis=[0, 0, 1], order=6),
               Reflection(label='sh', axis=[np.sqrt(2/9), 0, 1/3])],
              {'Ag': pd.Series([+1, +1, +1, +1, +1, +1]),
               'Eg': pd.Series([+2, -1, +2, +2, -1, +2]),
               'Tg': pd.Series([+3,  0, -1, +3,  0, -1]),
               'Au': pd.Series([+1, +1, +1, -1, -1, -1]),
               'Eu': pd.Series([+2, -1, +2, -2, +1, -2]),
               'Tu': pd.Series([+3,  0, -1, -3,  0, +1])
               },
              rotations=['Tg'],
              translations=['Tu'],
              multiplicities=[1, 8, 3, 1, 8, 3],
              generator_indices=[2, 4]),

    CharTable('O',
              [Identity(label='E'),
               Rotation(label='C3', axis=[1/np.sqrt(3), 1/np.sqrt(3), 1/np.sqrt(3)], order=3),
               Rotation(label='C2', axis=[0, 0, 1], order=2, exp=1),
               Rotation(label='C4', axis=[0, 0, 1], order=4),
               Rotation(label="C2'", axis=[1/np.sqrt(2), 1/np.sqrt(2), 0], order=2)],
              {'A1': pd.Series([+1, +1, +1, +1, +1]),
               'A2': pd.Series([+1, +1, +1, -1, -1]),
               'E' : pd.Series([+2, -1, +2,  0,  0]),
               'T1': pd.Series([+3,  0, -1, +1, -1]),
               'T2': pd.Series([+3,  0, -1, -1, +1])
               },
              rotations=['T1'],
              translations=['T1'],
              multiplicities=[1, 8, 3, 6, 6],
              generator_indices=[1, 3]),

    CharTable('Oh',
              [Identity(label='E'),
               Rotation(label='C3', axis=[1/np.sqrt(3), 1/np.sqrt(3), 1/np.sqrt(3)], order=3),
               Rotation(label='C2', axis=[0, 0, 1], order=2),
               Rotation(label='C4', axis=[0, 0, 1], order=4),
               Rotation(label="C2'", axis=[1/np.sqrt(2), 1/np.sqrt(2), 0], order=2),
               Inversion(label='i'),
               ImproperRotation(label='S6', axis=[1/np.sqrt(3), 1/np.sqrt(3), 1/np.sqrt(3)], order=6),
               Reflection(label='s_h', axis=[0, 0, 1]),
               ImproperRotation(label='S4', axis=[0, 0, 1], order=4),
               Reflection(label="s_d", axis=[1/np.sqrt(2), 1/np.sqrt(2), 0]),
               ],
              {'A1g': pd.Series([+1, +1, +1, +1, +1, +1, +1, +1, +1, +1]),
               'A2g': pd.Series([+1, +1, +1, -1, -1, +1, +1, +1, -1, -1]),
               'Eg' : pd.Series([+2, -1, +2,  0,  0, +2, -1, +2,  0,  0]),
               'T1g': pd.Series([+3,  0, -1, +1, -1, +3,  0, -1, +1, -1]),
               'T2g': pd.Series([+3,  0, -1, -1, +1, +3,  0, -1, -1, +1]),
               'A1u': pd.Series([+1, +1, +1, +1, +1, -1, -1, -1, -1, -1]),
               'A2u': pd.Series([+1, +1, +1, -1, -1, -1, -1, -1, +1, +1]),
               'Eu' : pd.Series([+2, -1, +2,  0,  0, -2, +1, -2,  0,  0]),
               'T1u': pd.Series([+3,  0, -1, +1, -1, -3,  0, +1, -1, +1]),
               'T2u': pd.Series([+3,  0, -1, -1, +1, -3,  0, +1, +1, -1]),
               },
              rotations=['T1g'],
              translations=['T1u'],
              multiplicities=[1, 8, 3, 6, 6, 1, 8, 3, 6, 6],
              generator_indices=[3, 6]),

    CharTable('I',
              [Identity(label='E'),
               Rotation(label='C5', axis=[0, -1, (np.sqrt(5) + 1)/2], order=5),
               Rotation(label='C^2_5', axis=[0, -1, (np.sqrt(5) + 1)/2], order=5, exp=2),
               Rotation(label='C3', axis=[1-(np.sqrt(5) + 1)/2, 0, (np.sqrt(5) + 1)/2], order=3),
               Rotation(label="C2'", axis=[0, 0, 1], order=2)],
              {'A' : pd.Series([+1, +1, +1, +1, +1]),
               'T1': pd.Series([+3,  real_radical(1, 10),  real_radical(3, 10),  0, -1]),
               'T2': pd.Series([+3,  real_radical(3, 10),  real_radical(1, 10),  0, -1]),
               'G' : pd.Series([+4, -1, -1, +1,  0]),
               'H' : pd.Series([+5,  0,  0, -1, +1])
               },
              rotations=['T1'],
              translations=['T1'],
              multiplicities=[1, 12, 12, 20, 15],
              generator_indices=[1, 3]),

    CharTable('Ih',
              [Identity(label='E'),
               Rotation(label='C5', axis=[0, -1, (np.sqrt(5) + 1) / 2], order=5),
               Rotation(label='C^2_5', axis=[0, -1, (np.sqrt(5) + 1) / 2], order=5, exp=2),
               Rotation(label='C3', axis=[1 - (np.sqrt(5) + 1) / 2, 0, (np.sqrt(5) + 1) / 2], order=3),
               Rotation(label="C2'", axis=[0, 0, 1], order=2),
               Inversion(label='i'),
               ImproperRotation(label='S_10', axis=[0, -1, (np.sqrt(5) + 1) / 2], order=10),
               ImproperRotation(label='S^3_10', axis=[0, -1, (np.sqrt(5) + 1) / 2], order=10, exp=3),
               ImproperRotation(label='S_6', axis=[1 - (np.sqrt(5) + 1) / 2, 0, (np.sqrt(5) + 1) / 2], order=6),
               Reflection(label='s_h', axis=[1, 0, 0])],
              {'Ag' : pd.Series([+1, +1, +1, +1, +1, +1, +1, +1, +1, +1]),
               'T1g': pd.Series([+3, real_radical(1, 10), real_radical(3, 10), 0, -1, +3, real_radical(3, 10), real_radical(1, 10), 0, -1,]),
               'T2g': pd.Series([+3, real_radical(3, 10), real_radical(1, 10), 0, -1, +3, real_radical(1, 10), real_radical(3, 10), 0, -1]),
               'Gg' : pd.Series([+4, -1, -1, +1, 0, +4, -1, -1, +1, 0]),
               'Hg' : pd.Series([+5, 0, 0, -1, +1, +5, 0, 0, -1, +1]),
               'Au' : pd.Series([+1, +1, +1, +1, +1, -1, -1, -1, -1, -1]),
               'T1u': pd.Series([+3, real_radical(1, 10), real_radical(3, 10), 0, -1, -3, -real_radical(3, 10), -real_radical(1, 10), 0, +1]),
               'T2u': pd.Series([+3, real_radical(3, 10), real_radical(1, 10), 0, -1, -3, -real_radical(1, 10), -real_radical(3, 10), 0, +1]),
               'Gu' : pd.Series([+4, -1, -1, +1,  0, -4, +1, +1, -1,  0]),
               'Hu' : pd.Series([+5,  0,  0, -1, +1, -5,  0,  0, +1, -1]),
               },
              rotations=['T1g'],
              translations=['T1u'],
              multiplicities=[1, 12, 12, 20, 15, 1, 12, 12, 20, 15],
              generator_indices=[1, 8]),
]

if __name__ == '__main__':


    for pg in ir_table_list[4:5]:
        print(pg.name)
        print(pg)
        print(len(pg.get_all_operations()))
        for op in pg.get_all_operations():
            print(len(pg.get_all_operations()[op]), op)

    exit()


    print(ir_table_list[-1].sort_index().index)
    # exit()

    a = np.array([ v for v in ir_table_list[-1].T.values]).T
    print('trans matrix')
    print(a)

    print('t1')
    t1 = np.dot(a, [0, 0, 0, 1, 0])
    print(t1)

    a_i = np.linalg.inv(a)
    print('inverse')
    print(a_i)

    print('t1(IR)')
    print(np.dot(a_i, t1))

    t1t1 = np.multiply(t1, t1)
    print('t1*t1', t1t1)

    e = np.dot(a, [0, 0, 1, 0, 0])
    et1 = np.multiply(e, t1)

    print('t1t1(IR)')
    print(np.dot(a_i, t1t1))
    print('et1')
    print(np.dot(a_i, et1))

    print('--------')

    td = ir_table_list[-1]

    print(0, 0, 0)
    print(0, 0, 1)
    print(np.sqrt(8 / 9), 0, -1 / 3)
    print(0, 1, 0)
    exit()

    # [Identity(label='E'), Rotation(label='C3', axis=[0, 0, 1], order=3),
    # Rotation(label='C2', axis=[np.sqrt(8 / 9), 0, -1 / 3], order=2),
    # ImproperRotation(label='S4', axis=[np.sqrt(8 / 9), 0, -1 / 3], order=4),



from dataclasses import dataclass


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super().__call__(*args, **kwargs)
        return cls._instances[cls]


@dataclass(frozen=False)
class Configuration(metaclass=Singleton):
    fast_optimization: bool = True  # speeds up the optimization by reducing the number of symmetry elements to use
    scan_steps: int = 20            # number of angle steps to use in the pre-scan in orientation optimization
    algorithm : str = 'hungarian'   # permutation algorithms: hungarian, exact


@dataclass(frozen=False)
class CustomPerm(metaclass=Singleton):
    perm_list: list = None



import numpy as np
from copy import deepcopy
import math
import itertools
from posym.integrals import product_poly_coeff, gaussian_integral, gaussian_integral_2 # noqa
from scipy.special import comb


def binomial_expansion(l, x, max_lim=None):

    if max_lim is None:
        max_lim = np.max(l)+1

    vector_x = np.zeros((max_lim, max_lim, max_lim))
    vector_y = np.zeros((max_lim, max_lim, max_lim))
    vector_z = np.zeros((max_lim, max_lim, max_lim))

    for k in range(l[0]+1):
        vector_x[k, 0, 0] += comb(l[0], k) * x[0]**(l[0]-k)
    for k in range(l[1] + 1):
        vector_y[0, k, 0] += comb(l[1], k) * x[1]**(l[1]-k)
    for k in range(l[2] + 1):
        vector_z[0, 0, k] += comb(l[2], k) * x[2]**(l[2]-k)

    vector = product_poly_coeff(product_poly_coeff(vector_x, vector_y), vector_z, max_lim)
    return vector


def product_poly_coeff_py(poly_coeff, poly_coeff2, max_lim=None):
    """
    Product of two polynomial coefficients matrices

    :param poly_coeff: matrix of polynomial coefficients 1
    :param poly_coeff2: matrix of polynomial coefficients 2
    :param max_lim:
    :return:
    """
    # max_lim = 0 if max_lim is None else max_lim
    # return product_poly_coeff_c(poly_coeff, poly_coeff2, max_lim)

    max_lim_1 = len(poly_coeff)
    max_lim_2 = len(poly_coeff2)
    max_lim_prod = max_lim_1 + max_lim_2
    poly_coeff_prod = np.zeros((max_lim_prod, max_lim_prod, max_lim_prod))

    for i, j, k in itertools.product(range(max_lim_1), repeat=3):
        for i2, j2, k2 in itertools.product(range(max_lim_2), repeat=3):
            poly_coeff_prod[i + i2, j + j2, k + k2] += poly_coeff[i, j, k] * poly_coeff2[i2, j2, k2]

    if max_lim is not None:
        poly_coeff_prod = poly_coeff_prod[:max_lim, :max_lim, :max_lim]

    return poly_coeff_prod


def exp_poly_coeff(poly_coeff, k, max_lim=None):
    """
    Calculate the exponential of a polynomial coefficient matrix

    :param poly_coeff: polynomial coefficient matrix
    :param k: exponent
    :param max_lim:
    :return:
    """

    max_lim_x = len(poly_coeff) * max(k, 1)

    poly_coeff_exp = np.zeros((max_lim_x, max_lim_x, max_lim_x))
    poly_coeff_exp[0, 0, 0] = 1
    for _ in range(k):
        poly_coeff_exp = product_poly_coeff(poly_coeff_exp, poly_coeff, max_lim=max_lim)

    return poly_coeff_exp


def gaussian_product(g_1, g_2):
    """
    Calculate the product of two Gaussian primitives. (s functions only)
    """
    alpha = g_1.alpha + g_2.alpha
    prefactor = g_1.prefactor * g_2.prefactor
    p = g_1.alpha * g_2.alpha

    PG = g_1.center - g_2.center
    coordinates = (g_1.center * g_1.alpha + g_2.center * g_2.alpha) / alpha
    K = np.exp(-p/alpha * np.dot(PG, PG))

    poly_coeff = product_poly_coeff(g_1.poly_coeff, g_2.poly_coeff)

    return PrimitiveGaussian(alpha, K * prefactor, coordinates, normalize=False, poly_coeff=poly_coeff)


def integrate_exponential_simple_py(n, a):
    """
    integrals type   x^n exp(-ax^2)

    """
    if n == 0:
        return np.sqrt(np.pi/a)
    elif np.mod(n, 2) == 0:
        k = n//2
        return 2*np.math.factorial(np.math.factorial(2*k-1))/(2**(k+1)*a**k)*np.sqrt(np.pi/a)
    else:
        return 0.0


def integrate_exponential_py(n, a, b):
    """
    integrals type  x^n exp(-(ax^2+bx))

    """
    if n == 0:
        return np.sqrt(np.pi/a)*np.exp(b**2/(4*a))
    elif n == 1:
        return np.sqrt(np.pi)/(2*a**(3/2))*b*np.exp(b**2/(4*a))
    else:
        factor = np.sum([comb(n, 2*k)*(b/(2*a))**(n-2*k)*math.factorial(2*k)/(2**(2*k)*math.factorial(k)*a**k)
                          for k in range(n//2+1)])
        return factor * np.sqrt(np.pi/a)*np.exp(b**2/(4*a))


def gaussian_integral_py(alpha, center, poly_coeff):
    max_lim = len(poly_coeff)
    pre_exponential = np.exp(-alpha * np.dot(center, center))

    integrate = 0.0
    for i, j, k in itertools.product(range(max_lim), repeat=3):
        integrate += poly_coeff[i, j, k] * np.prod([integrate_exponential_py(l, alpha, 2 * alpha * c) for c, l in zip(center, [i, j, k])])

    return pre_exponential * integrate


def integrate_exponential_sep_py(n, a, b):
    """
    norm for integrals type  x^n exp(-(ax^2+bx))
    separated in prefactor and exponential

    """
    if n == 0:
        return np.sqrt(np.pi/a), b**2/(4*a)
    elif n == 1:
        return np.sqrt(np.pi)/(2*a**(3/2))*b, b**2/(4*a)
    else:
        factor = np.sum([comb(n, 2*k)*(b/(2*a))**(n-2*k)*math.factorial(2*k)/(2**(2*k)*math.factorial(k)*a**k)
                          for k in range(n//2+1)])
        return factor * np.sqrt(np.pi/a), b**2/(4*a)


def gaussian_integral_2_py(alpha, center, poly_coeff):
    """
    integrals type  x^n exp(-(ax^2+bx))
    alternative implementation to handle large exponents

    """

    max_lim = len(poly_coeff)
    pre_exponent = -alpha * np.dot(center, center)

    exp_list = []
    pre_exp_list = []
    for i, j, k in itertools.product(range(max_lim), repeat=3):
        list_data = [integrate_exponential_sep_py(l, alpha, 2 * alpha * c) for c, l in zip(center, [i, j, k])]
        pre_exp_list.append(poly_coeff[i, j, k] * np.prod([l[0] for l in list_data]))
        exp_list.append(np.sum([l[1] for l in list_data]))

    lowest_index = np.argmin(exp_list)
    exponential = exp_list[lowest_index]
    exp_list = np.array(exp_list) - exponential

    pre_exponential = np.sum([a*np.exp(b) for a, b in zip(pre_exp_list, exp_list)])

    return pre_exponential * np.exp(exponential + pre_exponent)


def simplify_poly_coeff(poly_coeff):
    """
    Simplify a polynomial coefficient matrix

    :param poly_coeff: polynomial coefficient matrix
    :return: simplified polynomial coefficient matrix
    """

    n_dim = len(poly_coeff)
    k = 0
    for i in range(1, n_dim-1):

        if poly_coeff[-i, :, :].any() or poly_coeff[:, -i, :].any() or poly_coeff[:, :, -i].any():
            break
        k += 1

    if k == 0:
        return poly_coeff

    return poly_coeff[:-k, :-k, :-k]


class PrimitiveGaussian:
    def __init__(self, alpha, prefactor=1.0, center=(0, 0, 0), l=(0, 0, 0), normalize=True, poly_coeff=None):
        self._n_dim = len(center)
        self.alpha = alpha
        self.prefactor = prefactor
        self.center = np.array(center)
        if poly_coeff is None:
            self.poly_coeff = binomial_expansion(l, -self.center)
        else:
            self.poly_coeff = poly_coeff

        # normalize primitive such that <prim|prim> = 1
        if normalize:
            norm = self._get_norm()

            self.prefactor = prefactor / np.sqrt(norm)

        self.poly_coeff = simplify_poly_coeff(self.poly_coeff)

    def __hash__(self):
        return hash((self.prefactor, self.alpha, self.center.data.tobytes(), self.poly_coeff.data.tobytes()))

    def __eq__(self, other):
        return self.__hash__() == other.__hash__()

    @property
    def shape(self):
        return hash((self.prefactor, self.alpha, self.poly_coeff.data.tostring()))

    @property
    def integrate(self):
        return self.prefactor * gaussian_integral_2(self.alpha, self.center, self.poly_coeff)
        # return self.prefactor * gaussian_integral(self.alpha, self.center, self.poly_coeff)

    def _get_norm_l(self, l):

        def d_fact(n):
            #print(list(range(1, n, 2)))
            return np.prod(list(range(1, n, 2)))

        num = 2**np.sum(l)*self.alpha**((2*l[0] + 2*l[1] + 2*l[2]+ 3)/4)
        denom = d_fact(2*l[0]) * d_fact(2*l[1]) * d_fact(2*l[2])

        # print('res', (2*np.pi)**(3/4)* num/np.sqrt(denom))
        return (2*np.pi)**(3/4) * num/np.sqrt(denom)

    def _get_norm(self):
        poly_coeff_sq = product_poly_coeff(self.poly_coeff, self.poly_coeff)
        return self.prefactor * gaussian_integral_2(2 * self.alpha, self.center, poly_coeff_sq)
        # return self.prefactor * gaussian_integral(2 * self.alpha, self.center, poly_coeff_sq)


    def __call__(self, *value):
        value = np.array(value)
        max_lim = len(self.poly_coeff)

        # to work with vectorized like calls
        if len(value[0].shape) or len(value[1].shape) or len(value[2].shape) > 1:

            angular = np.zeros_like(value[0])
            radial = np.zeros_like(value[0])

            for indices in itertools.product(*[range(m) for m in value[0].shape]):
                coef_matrix = np.fromfunction(lambda i, j, k: value[0][indices] ** i * value[1][indices] ** j * value[2][indices] ** k,
                                              (max_lim, max_lim, max_lim))
                angular[indices] = np.sum(self.poly_coeff * coef_matrix)

                dot = np.sum([(value[m][indices] - self.center[m]) ** 2 for m in range(3)])
                radial[indices] = np.exp(-self.alpha * dot)

        else:
            coef_matrix = np.fromfunction(lambda i, j, k: value[0]**i*value[1]**j*value[2]**k, (max_lim, max_lim, max_lim))
            radial = np.exp(-self.alpha * np.linalg.norm(value - self.center) ** 2)
            angular = np.sum(self.poly_coeff * coef_matrix)

        return self.prefactor * angular * radial

    def __mul__(self, other):
        return gaussian_product(self, other)

    def apply_translation(self, translation):
        translation = np.array(translation)
        max_lim = len(self.poly_coeff)
        poly_coeff_trans = np.zeros_like(self.poly_coeff)

        for i, j, k in itertools.product(range(max_lim), repeat=3):
            poly_coeff_trans += self.poly_coeff[i, j, k] * binomial_expansion([i, j, k], -translation, max_lim=max_lim)
        self.center = self.center + translation
        self.poly_coeff = simplify_poly_coeff(poly_coeff_trans)
        return self

    def apply_linear_transformation(self, operation_matrix):

        max_lim = len(self.poly_coeff)

        if max_lim > 1:
            max_lim_final = max_lim * 3 # * NDIM (naive max dimmension can be improved)
            poly_coeff_rot = np.zeros((max_lim_final, max_lim_final, max_lim_final))

            for i, j, k in itertools.product(range(max_lim), repeat=3):
                poly_temp_xyz = np.zeros((max_lim_final, max_lim_final, max_lim_final))
                poly_temp_xyz[0, 0, 0] = 1.0
                for l, m in enumerate([i, j, k]):
                    poly_temp = np.zeros((max_lim, max_lim, max_lim))
                    poly_temp[1, 0, 0] = operation_matrix[0, l]
                    poly_temp[0, 1, 0] = operation_matrix[1, l]
                    poly_temp[0, 0, 1] = operation_matrix[2, l]
                    poly_temp = exp_poly_coeff(poly_temp, m, max_lim_final)

                    poly_temp_xyz = product_poly_coeff(poly_temp_xyz, poly_temp, max_lim_final)

                poly_coeff_rot += self.poly_coeff[i, j, k] * poly_temp_xyz

            self.poly_coeff = simplify_poly_coeff(poly_coeff_rot)
        self.center = np.dot(operation_matrix, self.center)
        return self

    def apply_rotation(self, angle, axis):
        """
        apply a rotation with respect to an axis to the function

        :param angle: angle in Radians
        :param axis: rotation axis
        :return:
        """
        from scipy.spatial.transform import Rotation as R

        rotation_vector = angle * np.array(axis) / np.linalg.norm(axis)
        rotation = R.from_rotvec(rotation_vector)

        self.apply_linear_transformation(rotation.as_matrix())
        return self

    def copy(self):
        return deepcopy(self)


class BasisFunction:
    def __init__(self, primitive_gaussians, coefficients, center=None, label=None):
        primitive_gaussians = deepcopy(primitive_gaussians)
        if center is not None:
            center = np.array(center)
            for primitive in primitive_gaussians:
                primitive.apply_translation(center - primitive.center)

        self.primitive_gaussians = primitive_gaussians
        self.coefficients = list(coefficients)
        self.label = label

    def get_number_of_primitives(self):
        return len(self.primitive_gaussians)

    def set_coordinates(self, coordinates):
        for primitive in self.primitive_gaussians:
            primitive.center = np.array(coordinates)

    def apply_translation(self, translation):
        primitive_gaussians = deepcopy(self.primitive_gaussians)
        for primitive in primitive_gaussians:
            primitive.apply_translation(translation)

        self.primitive_gaussians = primitive_gaussians
        return self

    def apply_linear_transformation(self, transformation_matrix):
        primitive_gaussians = deepcopy(self.primitive_gaussians)
        for primitive in primitive_gaussians:
            primitive.apply_linear_transformation(transformation_matrix)

        self.primitive_gaussians = primitive_gaussians
        return self

    def apply_rotation(self, angle, axis):
        """
        apply a rotation with respect to an axis to the function
        :param angle: angle in Radians
        :param axis: the axis
        :return:
        """
        primitive_gaussians = deepcopy(self.primitive_gaussians)
        for primitive in primitive_gaussians:
            primitive.apply_rotation(angle, axis)

        self.primitive_gaussians = primitive_gaussians
        return self

    def copy(self):
        return deepcopy(self)

    @property
    def integrate(self):
        return np.sum([coef * prim.integrate for coef, prim in zip(self.coefficients, self.primitive_gaussians)])

    def get_environment_centers(self):
        """
        Get an estimation of the molecular geometry by checking the centers of gaussian functions.
        Different environments (atom types) are distinguished as different gaussian exponents
        :return: symbols (environments types) and coordinates (gaussian centers)
        """

        unique_center = np.unique(np.array([p.center for p in self.primitive_gaussians]), axis=0)
        data_dict = {}
        coor_dict = {}

        for i, center in enumerate(unique_center):
            data = np.where((np.array([p.center for p in self.primitive_gaussians] == center).all(axis=1)))
            key = tuple(np.array([p.alpha for p in self.primitive_gaussians])[data])
            if key in data_dict:
                data_dict[key].append(i)
                coor_dict[key].append(center)
            else:
                data_dict[key] = [i]
                coor_dict[key] = [center]

        symbols_ = []
        coordinates_ = []
        for i, (key, value) in enumerate(data_dict.items()):
            for v, c in zip(value, coor_dict[key]):
                symbols_.append('{}'.format(i))
                coordinates_.append(list(c))

        return symbols_, coordinates_

    def global_center(self):
        centers = []
        weights = []
        for coef, prim in zip(self.coefficients, self.primitive_gaussians):
            centers.append(prim.center)
            weights.append(coef**2)

        return np.average(centers, weights=weights, axis=0)
    def __repr__(self):
        if self.label is not None:
            return '{}'.format(self.label)
        return super(BasisFunction, self).__repr__()

    def __call__(self, *value):
        return sum([coef * prim(*value) for coef, prim in zip(self.coefficients, self.primitive_gaussians)])

    def __mul__(self, other):
        if isinstance(other, (float, int)):
            return BasisFunction(self.primitive_gaussians, [coef * other for coef in self.coefficients])

        elif isinstance(other, BasisFunction):

            primitive_gaussians = []
            coefficients = []

            for primitive_1, coeff_1 in zip(self.primitive_gaussians, self.coefficients):
                for primitive_2, coeff_2 in zip(other.primitive_gaussians, other.coefficients):
                    primitive_gaussians.append(primitive_1 * primitive_2)
                    coefficients.append(coeff_1 * coeff_2)

            return BasisFunction(primitive_gaussians, coefficients)
        raise Exception("No product defined for type: {}".format(type(other)))

    def __rmul__(self, other):
        return self.__mul__(other)

    def __add__(self, other):

        primitive_gaussians = deepcopy(self.primitive_gaussians)
        coefficients = deepcopy(self.coefficients)

        for prim, coeff in zip(other.primitive_gaussians, other.coefficients):
            if prim in primitive_gaussians:
                coefficients[primitive_gaussians.index(prim)] += coeff
            else:
                primitive_gaussians.append(prim)
                coefficients.append(coeff)

        return BasisFunction(primitive_gaussians, coefficients)

    def __sub__(self, other):
        primitive_gaussians = deepcopy(self.primitive_gaussians)
        coefficients = deepcopy(self.coefficients)

        negative_coefficients = list(-np.array(other.coefficients))
        for prim, coeff in zip(other.primitive_gaussians, negative_coefficients):
            if prim in primitive_gaussians:
                coefficients[primitive_gaussians.index(prim)] += coeff
            else:
                primitive_gaussians.append(prim)
                coefficients.append(coeff)

        return BasisFunction(primitive_gaussians, coefficients)


if __name__ == '__main__':

    # test Lithium
    """
    Li     0
    S    3   1.00
          0.1611957475D+02       0.1543289673D+00
          0.2936200663D+01       0.5353281423D+00
          0.7946504870D+00       0.4446345422D+00
    SP   3   1.00
          0.6362897469D+00      -0.9996722919D-01       0.1559162750D+00
          0.1478600533D+00       0.3995128261D+00       0.6076837186D+00
          0.4808867840D-01       0.7001154689D+00       0.3919573931D+00
    ****
3   d
      0.2145684671D+02       0.2197679508D+00
      0.6545022156D+01       0.6555473627D+00
      0.2525273021D+01       0.2865732590D+00
*
    """

    sa = PrimitiveGaussian(alpha=16.11957475, l=[0, 0, 0], center=(0.0, 0.0, 0.0), normalize=True)
    sb = PrimitiveGaussian(alpha=2.936200663, l=[0, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    sc = PrimitiveGaussian(alpha=0.794650487, l=[0, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    s1 = BasisFunction([sa, sb, sc], [0.1543289673, 0.5353281423, 0.4446345422])
    print('s:', (s1*s1).integrate)
    s1.apply_rotation(0.33*2*np.pi/2, [0.2, 5.0, 1.0])

    s2a = PrimitiveGaussian(alpha=0.6362897469, l=[0, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    s2b = PrimitiveGaussian(alpha=0.1478600533, l=[0, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    s2c = PrimitiveGaussian(alpha=0.0480886784, l=[0, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    s2 = BasisFunction([s2a, s2b, s2c], [-0.09996722919, 0.3995128261, 0.7001154689])
    print('s2:', (s2*s2).integrate)

    pxa = PrimitiveGaussian(alpha=0.6362897469, l=[1, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    pxb = PrimitiveGaussian(alpha=0.1478600533, l=[1, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    pxc = PrimitiveGaussian(alpha=0.0480886784, l=[1, 0, 0], center=[0.0, 0.0, 0.0], normalize=True)
    px = BasisFunction([pxa, pxb, pxc], [0.1559162750, 0.6076837186, 0.3919573931],
                       center=[0.0, 0.0, 0.0]
                       )
    print('px:', (px*px).integrate)
    px.apply_rotation(np.pi/2, [0.0, 0.0, 1.0])

    print('px rot:', (px*px).integrate)

    print('pxa:', (pxa*pxa).integrate)
    pxa.apply_rotation(2*np.pi/2, [0.0, 0.0, 1.0])
    print('pxa:', (pxa*pxa).integrate)


    o1 = -0.992527759 * s1 -0.0293095626 * s2
    print('o1:', (o1*o1).integrate)

    o2 = 0.276812804 * s1 -1.02998914 * s2
    print('o2:', (o2*o2).integrate)

    pya = PrimitiveGaussian(alpha=0.6362897469, l=[0, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    pyb = PrimitiveGaussian(alpha=0.1478600533, l=[0, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    pyc = PrimitiveGaussian(alpha=0.0480886784, l=[0, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    py = BasisFunction([pya, pyb, pyc], [0.1559162750, 0.6076837186, 0.3919573931])
    print('py:', (py*py).integrate)

    pza = PrimitiveGaussian(alpha=0.6362897469, l=[0, 0, 1], center=[0.0, 0.0, 0.0], normalize=True)
    pzb = PrimitiveGaussian(alpha=0.1478600533, l=[0, 0, 1], center=[0.0, 0.0, 0.0], normalize=True)
    pzc = PrimitiveGaussian(alpha=0.0480886784, l=[0, 0, 1], center=[0.0, 0.0, 0.0], normalize=True)
    pz = BasisFunction([pza, pzb, pzc], [0.1559162750, 0.6076837186, 0.3919573931])
    print('pz:', (pz*pz).integrate)

    d1a = PrimitiveGaussian(alpha=21.45684671, l=[1, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    d1b = PrimitiveGaussian(alpha=6.545022156, l=[1, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    d1c = PrimitiveGaussian(alpha=2.525273021, l=[1, 1, 0], center=[0.0, 0.0, 0.0], normalize=True)
    d1 = BasisFunction([d1a, d1b, d1c], [0.2197679508, 0.6555473627, 0.2865732590])
    d1.apply_rotation(np.pi/4, [0.0, 0.0, 1.0]).apply_translation([0.2, 0, 0])
    print('d1:', (d1*d1).integrate)

    px2 = px * px
    print('px2:', px2.integrate)

    from scipy import integrate
    # num_integral = integrate.tplquad(px2, -5, 5, lambda x: -5, lambda x: 5, lambda x, y: -5, lambda x, y: 5)
    # print('num_integral px*px', num_integral)

    import matplotlib.pyplot as plt
    x = np.linspace(-2, 2, 100)
    y = np.linspace(-2, 2, 100)

    X, Y = np.meshgrid(x, y)
    Z = d1(X, Y, np.zeros_like(X))
    plt.imshow(Z, interpolation='bilinear', origin='lower', cmap='seismic')
    plt.figure()
    plt.contour(X, Y, Z, colors='k')
    plt.show()

    x = np.linspace(-5, 5, 200)
    zeros = np.zeros_like(x)
    plt.plot(x, o1(x, zeros, zeros), label='O1')
    plt.plot(x, o2(x,  zeros, zeros), label='O2')
    plt.plot(x, [px(x_,  0.0, 0.0)*py(x_, 0, 0)for x_ in x], label='px*py')
    plt.plot(x, [px2(x_, 0.0, 0.0) for x_ in x], '--', label='px2')
    plt.plot(x, [d1(x_,  0.0, 0.0) for x_ in x], '-', label='d1')
    plt.plot(x, [pxa(x_, 0.0, 0.0) for x_ in x], '-', label='pxa')

    plt.legend()
    plt.show()


    basis = {'name': 'STO-3G',
             'primitive_type': 'gaussian',
             'atoms': [{'symbol': 'O',
                        'shells': [{'shell_type': 's',
                                    'p_exponents': [130.70932, 23.808861, 6.4436083],
                                    'con_coefficients': [0.154328969, 0.535328136, 0.444634536],
                                    'p_con_coefficients': [0.0, 0.0, 0.0]},
                                   {'shell_type': 'sp',
                                    'p_exponents': [5.0331513, 1.1695961, 0.380389],
                                    'con_coefficients': [-0.0999672287, 0.399512825, 0.700115461],
                                    'p_con_coefficients': [0.155916268, 0.607683714, 0.391957386]}]},
                       {'symbol': 'H',
                        'shells': [{'shell_type': 's',
                                    'p_exponents': [3.42525091, 0.62391373, 0.1688554],
                                    'con_coefficients': [0.154328971, 0.535328142, 0.444634542],
                                    'p_con_coefficients': [0.0, 0.0, 0.0]}]},
                       {'symbol': 'H',
                        'shells': [{'shell_type': 's',
                                    'p_exponents': [3.42525091, 0.62391373, 0.1688554],
                                    'con_coefficients': [0.154328971, 0.535328142, 0.444634542],
                                    'p_con_coefficients': [0.0, 0.0, 0.0]}]}]}

    mo_coefficients = [[ 0.994216442, 0.025846814, 0.000000000, 0.000000000,-0.004164076,-0.005583712, -0.005583712],
                       [ 0.233766661,-0.844456594, 0.000000000, 0.000000000, 0.122829781,-0.155593214, -0.155593214],
                       [ 0.000000000, 0.000000000, 0.612692349, 0.000000000, 0.000000000,-0.449221684,  0.449221684],
                       [-0.104033343, 0.538153649, 0.000000000, 0.000000000, 0.755880259,-0.295107107, -0.295107107],
                       [ 0.000000000, 0.000000000, 0.000000000,-1.000000000, 0.000000000, 0.000000000,  0.000000000],
                       [-0.125818566, 0.820120983, 0.000000000, 0.000000000,-0.763538862,-0.769155124, -0.769155124],
                       [ 0.000000000, 0.000000000, 0.959800163, 0.000000000, 0.000000000, 0.814629717, -0.814629717]]

    coordinates=[[ 0.0000000000, 0.000000000, -0.0428008531],
                 [-0.7581074140, 0.000000000, -0.6785995734],
                 [ 0.7581074140, 0.000000000, -0.6785995734]]

    # Oxigen atom
    sa = PrimitiveGaussian(alpha=130.70932)
    sb = PrimitiveGaussian(alpha=23.808861)
    sc = PrimitiveGaussian(alpha=6.4436083)
    s_O = BasisFunction([sa, sb, sc],
                        [0.154328969, 0.535328136, 0.444634536],
                        center=[0.0000000000, 0.000000000, -0.0808819]) # Bohr

    sa = PrimitiveGaussian(alpha=5.03315132)
    sb = PrimitiveGaussian(alpha=1.1695961)
    sc = PrimitiveGaussian(alpha=0.3803890)
    s2_O = BasisFunction([sa, sb, sc],
                         [-0.099967228, 0.399512825, 0.700115461],
                         center=[0.0000000000, 0.000000000, -0.0808819])

    pxa = PrimitiveGaussian(alpha=5.0331513, l=[1, 0, 0])
    pxb = PrimitiveGaussian(alpha=1.1695961, l=[1, 0, 0])
    pxc = PrimitiveGaussian(alpha=0.3803890, l=[1, 0, 0])

    pya = PrimitiveGaussian(alpha=5.0331513, l=[0, 1, 0])
    pyb = PrimitiveGaussian(alpha=1.1695961, l=[0, 1, 0])
    pyc = PrimitiveGaussian(alpha=0.3803890, l=[0, 1, 0])

    pza = PrimitiveGaussian(alpha=5.0331513, l=[0, 0, 1])
    pzb = PrimitiveGaussian(alpha=1.1695961, l=[0, 0, 1])
    pzc = PrimitiveGaussian(alpha=0.3803890, l=[0, 0, 1])

    px_O = BasisFunction([pxa, pxb, pxc],
                         [0.155916268, 0.6076837186, 0.3919573931],
                         center=[0.0000000000, 0.000000000, -0.0808819])
    py_O = BasisFunction([pya, pyb, pyc],
                         [0.155916268, 0.6076837186, 0.3919573931],
                         center=[0.0000000000, 0.000000000, -0.0808819])
    pz_O = BasisFunction([pza, pzb, pzc],
                         [0.155916268, 0.6076837186, 0.3919573931],
                         center=[0.0000000000, 0.000000000, -0.0808819])

    # Hydrogen atoms
    sa = PrimitiveGaussian(alpha=3.42525091)
    sb = PrimitiveGaussian(alpha=0.62391373)
    sc = PrimitiveGaussian(alpha=0.1688554)
    s_H = BasisFunction([sa, sb, sc],
                        [0.154328971, 0.535328142, 0.444634542],
                        center=[-1.43262, 0.000000000, -1.28237])

    s2_H = BasisFunction([sa, sb, sc],
                         [0.154328971, 0.535328142, 0.444634542],
                         center=[1.43262, 0.000000000, -1.28237])

    o1 = s_O * 0.994216442 + s2_O * 0.025846814 + px_O * 0.0 + py_O * 0.0 + pz_O * \
         -0.004164076 + s_H * -0.005583712 + s2_H * -0.005583712

    o2 = s_O * 0.0 + s2_O * 0.0 + px_O * 0.612692349 + py_O * 0.0 + pz_O * 0.0 + s_H * \
         -0.44922168 + s2_H * 0.449221684


    x = np.linspace(-2, 2, 50)
    y = np.linspace(-2, 2, 50)

    X, Y = np.meshgrid(x, y)
    o2.apply_rotation(np.pi/2, [1, 0, 0])
    o2.apply_translation([0, -0.5, 0])
    Z = o2(X, Y, np.zeros_like(X))
    plt.imshow(Z, interpolation='bilinear', origin='lower', cmap='seismic')
    plt.figure()
    plt.contour(X, Y, Z, colors='k')
    plt.show()

    print('dot o1o1', (o1*o1).integrate)
    print('dot o2o2', (o2*o2).integrate)
    print('dot o1o2', (o2*o1).integrate)

    density_matrix = 0 * np.outer(mo_coefficients[0], mo_coefficients[0]) + \
                     0 * np.outer(mo_coefficients[1], mo_coefficients[1]) + \
                     2 * np.outer(mo_coefficients[2], mo_coefficients[2]) + \
                     2 * np.outer(mo_coefficients[3], mo_coefficients[3]) + \
                     0 * np.outer(mo_coefficients[4], mo_coefficients[4]) + \
                     0 * np.outer(mo_coefficients[5], mo_coefficients[5]) + \
                     0 * np.outer(mo_coefficients[6], mo_coefficients[6])

    print('density matrix\n', density_matrix[:4, :4])

    basis_functions = [s_O, s2_O, px_O, py_O, pz_O, s_H, s2_H]
    total_electrons = 0
    for i, bf1 in enumerate(basis_functions):
        for j, bf2 in enumerate(basis_functions):
            total_electrons += (bf1*bf2).integrate * density_matrix[i, j]


    print('total electrons', total_electrons)

    from posym.tools import rotate_basis_set, translate_basis_set, get_self_similarity, build_density


    density = build_density(basis_functions, density_matrix)
    density.apply_rotation(-np.pi/2, [1, 0, 0])
    density.apply_translation([0, 0.0, 0])

    print('density (valence)\n', density(0, 0, 0))
    print('n_electrons (valence)\n', density.integrate)
    print('self_similarity (valence)\n', (density*density).integrate)

    X, Y = np.meshgrid(x, y)
    # density.apply_rotation(np.pi/4, [1, 0, 0])
    Z = density(X, Y, np.zeros_like(X))
    plt.imshow(Z, interpolation='bilinear', origin='lower', cmap='seismic')
    plt.figure()
    plt.contour(X, Y, Z, colors='k')
    plt.show()

    # self_similarity = get_overlap_density(basis_functions, basis_functions, density_matrix)
    # print('self_similarity', self_similarity)

    self_similarity = get_self_similarity(basis_functions, density_matrix)
    print('self_similarity', self_similarity)

    basis_functions_r = rotate_basis_set(basis_functions, np.pi, [1, 0, 0])
    basis_functions_t = translate_basis_set(basis_functions, [1, 0, 0])

    s_O, s2_O, px_O, py_O, pz_O, s_H, s2_H = basis_functions_r
    o2 = s_O * 0.0 + s2_O * 0.0 + px_O * 0.612692349 + py_O * 0.0 + pz_O * 0.0 + s_H * -0.44922168 + s2_H * 0.449221684
    print('dot(rot)', (o2*o2).integrate)

    s_O, s2_O, px_O, py_O, pz_O, s_H, s2_H = basis_functions_t
    o2 = s_O * 0.0 + s2_O * 0.0 + px_O * 0.612692349 + py_O * 0.0 + pz_O * 0.0 + s_H * -0.44922168 + s2_H * 0.449221684
    print('dot(trans)', (o2*o2).integrate)

    # print('measure dens:', get_overlap_density(basis_functions, basis_functions_r, density_matrix)/self_similarity)




from functools import lru_cache
# from posym.permutations import get_cross_distance_table
from scipy.optimize import linear_sum_assignment
import numpy as np


@lru_cache(maxsize=100)
def get_submatrix_indices(symbols):
    # separate distance_table in submatrices corresponding to a single symbol
    submatrices_indices = []
    for s in np.unique(symbols):
        submatrices_indices.append([j for j, s2 in enumerate(symbols) if s2 == s])

    return submatrices_indices


def get_permutation_labels(distance_table, symbols, permutation_function):
    """
    This function restricts permutations by the use of atom labels
    returns the permutation vector that minimizes its trace using custom algorithms.
    """
    submatrices_indices = get_submatrix_indices(symbols)

    # determine the permutation for each submatrix
    perm_submatrices = []
    for index in submatrices_indices:
        submatrix = np.array(distance_table)[index, :][:, index]
        perm_sub = permutation_function(submatrix)
        perm_submatrices.append(perm_sub)

    # restore global permutation by joining permutations of submatrices
    global_permutation = np.zeros(len(distance_table), dtype=int)
    for index, perm in zip(submatrices_indices, perm_submatrices):
        index = np.array(index)
        global_permutation[index] = index[perm]

    return np.array(global_permutation)


def cache_permutation(func):
    cache_dict = {}

    def wrapper_cache(operation, coordinates, symbols):
        hash_key = (np.array2string(operation), np.array2string(coordinates), tuple(symbols))
        if hash_key in cache_dict:
            return cache_dict[hash_key]

        cache_dict[hash_key] = func(operation, coordinates, symbols)
        return cache_dict[hash_key]

    return wrapper_cache


@cache_permutation
def get_permutation_hungarian(operation, coordinates, symbols):

    operated_coor = np.dot(operation, coordinates.T).T
    symbols = tuple(int.from_bytes(num.encode(), 'big') for num in symbols)

    dot_table = -np.dot(coordinates, operated_coor.T)
    # dot_table = get_cross_distance_table(coordinates, operated_coor)

    # permutation algorithms functions
    def hungarian_algorithm(sub_matrix):
        row_ind, col_ind = linear_sum_assignment(sub_matrix)
        perm = np.zeros_like(row_ind)
        perm[row_ind] = col_ind
        return perm

    return get_permutation_labels(dot_table, symbols, hungarian_algorithm)



from posym.operations.identity import Identity
from posym.operations.reflection import Reflection
from posym.operations.rotation import Rotation
from posym.operations.irotation import ImproperRotation
from posym.operations.inversion import Inversion
from fractions import Fraction
import numpy as np


def test_matrix(matrix1, matrix2):
    if not np.allclose(matrix1, matrix2):
        raise Exception('eps!')


def get_degeneracy(eval, tol=1e-5):

    eval = np.real(eval)

    n_ones = 0
    n_iones = 0
    i_one = None
    i_ione = None

    for i, val in enumerate(eval):
        if 1 - tol < val < 1 + tol:
            n_ones += 1
            i_one = i
        if -1 - tol < val < -1 + tol:
            n_iones += 1
            i_ione = i

    if n_ones == 3:
        return 'E', None
    if n_iones == 3:
        return 'i', None
    if pow(-1, n_iones) == 1:
        return 'r', i_one
    if pow(-1, n_iones) == -1:
        return 's', i_ione

    raise Exception('Operation not recognized')


def get_operation_from_matrix(matrix, tol=1e-5):

    eval, evec = np.linalg.eig(matrix)
    op_type, index = get_degeneracy(eval)

    if op_type == 'E':
        return Identity(label='*E')

    if op_type == 'i':
        return Inversion(label='*i')

    if op_type == 'r':
        axis = np.real(evec.T[index])
        out_axis = [i for i in range(3) if i != index]

        sign = np.sign(np.dot(np.cross(evec.T[out_axis[0]], evec.T[out_axis[1]]), axis).imag)
        sign = sign if sign != 0 else 1

        index_2 = 0 if index != 0 else 1

        # handle numerical error
        eval_2 = np.real(eval[index_2])
        if eval_2 > 1: eval_2 -= 2
        if eval_2 < -1: eval_2 += 2

        angle = np.arccos(eval_2)

        if abs(angle) < tol:
            return Rotation('*R2', axis, order=2, exp=1)

        f = Fraction(2*np.pi/angle).limit_denominator(100)
        order = int(f.numerator)
        exp = int(sign * f.denominator)

        return Rotation('*R{}'.format(order), axis, order=order, exp=exp)

    if op_type == 's':
        axis = np.real(evec.T[index])
        out_axis = [i for i in range(3) if i != index]

        sign = np.sign(np.dot(np.cross(evec.T[out_axis[0]], evec.T[out_axis[1]]), axis).imag)
        sign = sign if sign != 0 else 1

        index_2 = 0 if index != 0 else 1

        if abs(np.real(eval[index_2])-1) < tol:
            test_matrix(Reflection(label='*M', axis=axis).matrix_representation, matrix)
            return Reflection(label='*M', axis=axis)

        # handle numerical error
        eval_2 = np.real(eval[index_2])

        if eval_2 > 1: eval_2 -= 2
        if eval_2 < -1: eval_2 += 2

        angle = np.arccos(eval_2)

        if abs(angle) < tol:
            return Reflection(label='*M', axis=axis)

        f = Fraction(2*np.pi/angle).limit_denominator(5)
        order = int(f.numerator)
        exp = int(sign * f.denominator)

        return ImproperRotation('*S{}'.format(order), axis, order=order, exp=exp)

    raise Exception('Operation not recognized')


def get_operation_from_matrix_new(matrix, tol=1e-5):

    eval, evec = np.linalg.eig(matrix)
    determinat = np.linalg.det(matrix)

    # determine degeneracies and main axis indices
    op_type, index = get_degeneracy(eval)

    if index is None:
        if determinat < 0:
            return Identity(label='*E')
        else:
            return Inversion(label='*I')

    # main axis
    main_axis = np.real(evec.T[index])
    side_axis = [i for i in range(3) if i != index]

    sign = np.sign(np.dot(np.cross(evec.T[side_axis[0]], evec.T[side_axis[1]]), main_axis).imag)
    sign = sign if sign != 0 else 1

    # chose a side axis to get the angle
    index_2 = 0 if index != 0 else 1

    # handle numerical error
    eval_2 = np.real(eval[index_2])
    if eval_2 > 1: eval_2 -= 2
    if eval_2 < -1: eval_2 += 2

    # get angle
    angle = np.arccos(eval_2)

    # rotation route
    if determinat > 0:

        if abs(angle) < tol:
            return Rotation('*R2', main_axis, order=2, exp=1)

        f = Fraction(2*np.pi/angle).limit_denominator(100)
        order = int(f.numerator)
        exp = int(sign * f.denominator)

        return Rotation('*R{}'.format(order), main_axis, order=order, exp=exp)

    # improper rotation route
    if determinat < 0:

        if abs(angle) < tol:
            return Reflection(label='*M', axis=main_axis)

        f = Fraction(2*np.pi/angle).limit_denominator(5)
        order = int(f.numerator)
        exp = int(sign * f.denominator)

        return ImproperRotation('*S{}'.format(order), main_axis, order=order, exp=exp)

    raise Exception('Operation not recognized')


def get_operation_from_matrix_test(matrix, tol=1e-5):

    eval, evec = np.linalg.eig(matrix)
    op_type, index = get_degeneracy(eval)

    if op_type == 'E':
        return Identity(label='*E')

    if op_type == 'i':
        return Inversion(label='*i')

    if op_type == 'r':
        axis = np.real(evec.T[index])
        out_axis = [i for i in range(3) if i != index]

        # print(np.dot(np.cross(evec.T[out_axis[0]], evec.T[out_axis[1]]), axis))
        sign = np.sign(np.dot(np.cross(evec.T[out_axis[0]], evec.T[out_axis[1]]), axis).imag)
        sign = sign if sign != 0 else 1

        index_2 = 0 if index != 0 else 1

        # handle numerical error
        eval_2 = np.real(eval[index_2])
        if eval_2 > 1: eval_2 -= 2
        if eval_2 < -1: eval_2 += 2

        angle = np.arccos(eval_2)

        if abs(angle) < tol:
            return Rotation('*R2', axis, order=2, exp=1)

        f = Fraction(2*np.pi/angle).limit_denominator(100)
        order = int(f.numerator)
        exp = int(sign * f.denominator)

        return Rotation('*R{}'.format(order), axis, order=order, exp=exp)

    if op_type == 's':
        axis = np.real(evec.T[index])
        out_axis = [i for i in range(3) if i != index]

        sign = np.sign(np.dot(np.cross(evec.T[out_axis[0]], evec.T[out_axis[1]]), axis).imag)
        sign = sign if sign != 0 else 1

        index_2 = 0 if index != 0 else 1

        if abs(np.real(eval[index_2])-1) < tol:
            test_matrix(Reflection(label='*M', axis=axis).matrix_representation, matrix)
            return Reflection(label='*M', axis=axis)

        # handle numerical error
        eval_2 = np.real(eval[index_2])

        if eval_2 > 1: eval_2 -= 2
        if eval_2 < -1: eval_2 += 2

        angle = np.arccos(eval_2)

        if abs(angle) < tol:
            return Reflection(label='*M', axis=axis)

        # print('angle: ', np.rad2deg(angle), (2*np.pi)/angle)
        # print('axis: ', axis)

        f = Fraction(2*np.pi/angle).limit_denominator(5)
        # print('Fraction: ', f, f.numerator, f.denominator)
        order = int(f.numerator)
        exp = int(sign * f.denominator)
        # print('order: ', order, 'exp: ', exp)

        return ImproperRotation('*S{}'.format(order), axis, order=order, exp=exp)

    raise Exception('Operation not recognized')



from posym.operations import Operation
import numpy as np


class Identity(Operation):
    def __init__(self, label):
        super().__init__(label)

    def __str__(self):
        return 'SymOp.Identity {} <{}>'.format(self._label, hex(id(self)))

    def get_measure_modes(self, modes, orientation=None):
        self._measure_mode = [1.0] * len(modes)

        return np.array(self._measure_mode)

    def get_measure_atom(self, orientation=None):
        return len(self.permutation)

    def get_measure_xyz(self, orientation=None):
        return 3

    def get_displacements_projection(self, orientation=None):
        n_atoms = len(self.permutation)
        return np.identity(3*n_atoms).reshape(3*n_atoms, n_atoms, 3)#.tolist()

    def get_operated_coordinates(self, coordinates, orientation=None):
        return np.array(coordinates)

    def get_overlap_func(self, op_function1, op_function2, orientation=None):
        return (op_function1*op_function2).integrate

    def get_measure_pos(self, coordinates, orientation=None, normalized=True):
        if normalized:
            return 1.0
        else:
            return np.einsum('ij, ij -> ', coordinates, coordinates)

    def apply_rotation(self, orientation):
        pass

    @property
    def matrix_representation(self):
        return np.identity(3)

    def __mul__(self, other):
        if not other.__class__.__bases__[0] is Operation:
            raise Exception('Product only defined between Operation subclasses')
        else:
            from posym.operations.products import get_operation_from_matrix
            matrix_product = self.matrix_representation @ other.matrix_representation
            return get_operation_from_matrix(matrix_product)




import numpy as np
from posym.permutation.permutations import validate_permutation  # noqa


class Operation:
    def __init__(self, label):
        self._label = label
        self._order = 1
        self._exp = 1
        self._determinant = 1
        self._gen_rep = []
        self._permutation = None
        self._hash = None

    def __hash__(self):
        if self._hash is None:
            vector = np.round(np.array(self.matrix_representation).flatten(), decimals=6)
            self._hash = hash(np.array(vector * 1e5, dtype=int).tobytes())

        return self._hash

    def __eq__(self, other):
        return hash(self) == hash(other)

    def get_type(self):
        normalized_exp = np.mod(self._exp, self._order)
        if normalized_exp > self._order//2:
            normalized_exp = self._order - normalized_exp

        return '{} : {} / {}'.format(type(self).__name__, abs(normalized_exp), self._order)

    def set_permutation(self, permutation):
        self._permutation = permutation

    def set_permutation_set(self, permutation_set, symbols, ignore_compatibility=False):

        def apply_op(permut, base):
            new = []
            for p, b in zip(permut, base):
                new.append(permut[b])
            return new

        n_atoms = len(symbols)
        permutation = list(range(n_atoms))
        for op in self._gen_rep:
            permutation = apply_op(permutation, permutation_set[op])

        if ignore_compatibility:
            self._permutation = permutation
            return True

        if validate_permutation(permutation, self._order, self._determinant):
            self._permutation = permutation
            return True

    def inverse(self):
        return self

    @property
    def permutation(self):
        if self._permutation is None:
            raise Exception('No permutation has been defined in', self)
        return self._permutation

    @property
    def label(self):
        return self._label

    @property
    def matrix_representation(self):
        raise NotImplementedError('Not implemented')

    def __mul__(self, other):
        if not other.__class__.__bases__[0] is Operation:
            raise Exception('Product only defined between Operation subclasses')
        else:
            from posym.operations.products import get_operation_from_matrix, get_operation_from_matrix_test
            matrix_product = self.matrix_representation @ other.matrix_representation

            new_operator = get_operation_from_matrix(matrix_product)
            if not np.allclose(new_operator.matrix_representation, matrix_product):
                print(self, ' * ',  other, ' = ', new_operator)
                print(new_operator.matrix_representation)
                print(matrix_product)
                get_operation_from_matrix_test(matrix_product)

                raise Exception('Product error!')

            new_operator._gen_rep = list(other._gen_rep) + list(self._gen_rep)
            return new_operator



from setuptools import setup, Extension
import numpy

include_dirs_numpy = [numpy.get_include()]

def get_version_number():
    main_ns = {}
    for line in open('posym/__init__.py', 'r').readlines():
        if not(line.find('__version__')):
            exec(line, main_ns)
            return main_ns['__version__']

def check_compiler():
    import subprocess
    output = subprocess.Popen(['gcc'], stderr=subprocess.PIPE).communicate()[1]
    if b'clang' in output:
        return 'clang'
    if b'gcc' in output:
        return 'gcc'


if check_compiler() == 'clang':
    integrals = Extension('posym.integrals',
                          extra_compile_args=['-std=c99'],
                          include_dirs=include_dirs_numpy,
                          sources=['c/integrals.c'])

else:
    print ('openmp is used')
    integrals = Extension('posym.integrals',
                          extra_compile_args=['-std=c99', '-fopenmp'],
                          extra_link_args=['-lgomp'],
                          include_dirs=include_dirs_numpy,
                          sources=['c/integrals.c'])

permutations = Extension('posym.permutation.permutations',
                         extra_compile_args=['-std=c99'],
                         include_dirs=include_dirs_numpy,
                         sources=['c/permutations.c'])


setup(name='posym',
      version=get_version_number(),
      description='posym module',
      long_description=open('README.md').read(),
      long_description_content_type='text/markdown',
      author='Abel Carreras',
      url='https://github.com/abelcarreras/posym',
      author_email='abelcarreras83@gmail.com',
      packages=['posym',
                'posym.operations',
                'posym.permutation'],
      install_requires=['numpy', 'scipy', 'pandas', 'PyYaml'],
      license='MIT License',
      ext_modules=[integrals, permutations])



import numpy as np
import copy


class Permutation:
    def __init__(self, permutation):
        self._permutation = copy.deepcopy(list(permutation))
        self._orbits = None

    def __str__(self):
        return self._permutation.__str__()

    def __iter__(self):
        return self._permutation.__iter__()

    def __getitem__(self, item):
        return self._permutation.__getitem__(item)

    def get_orbits(self):
        if self._orbits is None:
            self._orbits = []
            track_pos = []

            for p in self._permutation:

                if not p in track_pos:
                    track_pos.append(p)
                    orbit = [p]
                    while orbit[0] != self._permutation[p]:
                        p = self._permutation[p]
                        track_pos.append(p)
                        orbit.append(p)
                    self._orbits.append(orbit)

        return self._orbits

    def len_orbits(self):
        len_orbits = []
        for orbit in self.get_orbits():
            len_orbits.append(len(orbit))

        return len_orbits

    def max_orbit(self):
        max_orb = 0
        for orbit in self.get_orbits():
            max_orb = len(orbit) if len(orbit) > max_orb else max_orb

        return max_orb

    def slide(self, n):
        slide_permut = np.array(self._permutation)

        for orbit in self.get_orbits():
            original = list(orbit)
            roll_orbit = np.roll(orbit, -n)

            slide_permut[original] = slide_permut[roll_orbit]

        return slide_permut

    @property
    def raw_list(self):
        return self._permutation

    def __len__(self):
        return len(self._permutation)


def generate_permutation_set(generators, symbols):

    from itertools import permutations, product

    n_atoms = len(symbols)

    def symbols_compatible(len_orbits):
        val = 1
        for lo in len_orbits:
            val *= len(np.unique(np.array(symbols)[lo]))
        if val == 1:
            return True
        return False

    def gen_perm(gen):
        order = gen._order
        determinant = gen._determinant
        for p in permutations(range(n_atoms)):
            p_object = Permutation(p)
            len_orbits = p_object.len_orbits()
            orbits = p_object.get_orbits()

            # check symbols restriction
            if not symbols_compatible(orbits):
                continue

            orbit_mod = np.ones_like(len_orbits)
            orbit_mod = np.multiply(orbit_mod, np.mod(order, len_orbits))

            if determinant < 0:
                orbit_mod = np.multiply(orbit_mod, np.mod(2 * order, len_orbits))
                if order != 2:
                    orbit_mod = np.multiply(orbit_mod, np.mod(2, len_orbits))

            if np.sum(orbit_mod) == 0:
                yield p

    # for perm_set in product(*gen_perm_list):
    for perm_set in product(*[gen_perm(gen) for gen in generators]):
        yield {k: pi for k, pi in zip(generators, perm_set)}



import numpy as np
from posym.operations import Operation
from posym.tools import standardize_vector


def cache_reflection(func):
    cache_dict = {}

    def wrapper_cache(rotation_axis):
        hash_key = tuple(rotation_axis)
        if hash_key in cache_dict:
            return cache_dict[hash_key]

        cache_dict[hash_key] = func(rotation_axis)
        return cache_dict[hash_key]

    return wrapper_cache


@cache_reflection
def reflection(reflection_axis):
    uax = np.dot(reflection_axis, reflection_axis)
    return np.identity(3) - 2*np.outer(reflection_axis, reflection_axis)/uax


class Reflection(Operation):
    def __init__(self, label, axis):
        super().__init__(label)
        self._order = 1
        self._determinant = -1

        self._axis = standardize_vector(axis)

    def __str__(self):
        axis_txt = '[{:8.3f} {:8.3f} {:8.3f}]'.format(*self._axis)
        return 'SymOp.Reflection {} {} <{}>'.format(self._label, axis_txt, hex(id(self)))

    def get_measure_modes(self, modes, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = reflection(rotated_axis)

        measure_mode = []
        for i, mode in enumerate(modes):

            operated_mode = np.dot(operation, np.array(mode).T).T
            norm = np.linalg.norm(mode)
            permu_mode = operated_mode[self.permutation]

            measure_mode.append(np.trace(np.dot(mode, permu_mode.T))/norm)

        return np.array(measure_mode)

    def get_measure_atom(self):

        measure_atoms = np.array([1 if i == p else 0 for i, p in enumerate(self.permutation)])

        return np.sum(measure_atoms)

    def get_measure_xyz(self, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = reflection(rotated_axis)

        measure_mode = []
        for axis in [[1, 0, 0], [0, 1 ,0], [0, 0, 1]]:
            operated_axis = np.dot(operation, axis)
            measure_mode.append(np.dot(axis, operated_axis))

        return np.sum(measure_mode)

    def get_displacements_projection(self, orientation=None):

        n_atoms = len(self.permutation)
        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = reflection(rotated_axis)

        cartesian_modes = np.identity(3 * n_atoms).reshape(3 * n_atoms, n_atoms, 3)

        projected_modes = []
        for i, mode in enumerate(cartesian_modes):
            operated_mode = np.dot(operation, np.array(mode).T).T
            projected_modes.append(operated_mode[self.permutation])

        return np.array(projected_modes)

    def get_measure_pos(self, coordinates, orientation=None, normalized=True):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = reflection(rotated_axis)

        operated_coor = np.dot(operation, coordinates.T).T
        mesure_coor = np.einsum('ij, ij -> ', coordinates, operated_coor[self.permutation])

        if normalized:
            mesure_coor /= np.einsum('ij, ij -> ', coordinates, coordinates)

        return mesure_coor

    def get_operated_coordinates(self, coordinates, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = reflection(rotated_axis)

        #if self._permutation is None:
        #    from posym.operations import get_permutation_aprox
        #    self._permutation = get_permutation_aprox(operation, coordinates, symbols, self._order)

        operated_coor = np.dot(operation, coordinates.T).T
        return operated_coor[self.permutation]

    def get_overlap_func(self, op_function1, op_function2, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        operation = reflection(rotated_axis)

        fn_function_r = op_function1.copy()
        fn_function_r.apply_linear_transformation(operation)

        return (op_function2*fn_function_r).integrate

    def apply_rotation(self, orientation):
        self._axis = orientation.apply(self._axis)

    def inverse(self):
        return Reflection(self.label, axis=self._axis)

    @property
    def axis(self):
        return self._axis

    @property
    def matrix_representation(self):
        return reflection(self._axis)

if __name__ == '__main__':

    from posym.operations.rotation import rotation
    from posym.operations.inversion import inversion

    def d(a):
        # print('initial', np.array(a)/np.linalg.norm(a))
        return np.array(a)/np.linalg.norm(a)

    axis = d([0, 0, 1])
    matrix_1 = reflection(axis)
    print('m1\n', matrix_1)
    # exit()

    #matrix_2 = inversion()
    print('ini_axis: ', -d([0, 0, 1]))
    matrix_2 = rotation(2*np.pi/6*1, d([0, 0, 1]))
    print('m2\n', matrix_2)

    m12 = matrix_2 @ matrix_2 @ matrix_1 @ matrix_1

    m12 = m12
    print('m12')
    print(m12)

    from posym.operations.products import get_operation_from_matrix
    op = get_operation_from_matrix(m12)
    print('op: ', op)
    print(op.matrix_representation)

    print('check')
    print(np.round(op.matrix_representation - m12, decimals=3))

    exit()

    print('det', np.linalg.det(m12))
    eval, evec = np.linalg.eig(m12)
    print('eval', eval)
    print(evec.T)

    axis = np.real(evec.T[2])

    angle = np.arccos(np.real(eval[0]))
    print('angle: ', angle, 2*np.pi/3)

    exit()

    # m12_test = reflection(axis)
    m12_test = rotation(2*np.pi/3, axis)

    print('test')
    print(m12_test)
    print(m12_test - m12)







from posym.operations import Operation
import numpy as np


def prepare_vector(positions, vector):
    return np.array(vector) + np.array(positions)


def inversion():
    return -np.identity(3)


class Inversion(Operation):
    def __init__(self, label):
        super().__init__(label)
        self._order = 1
        self._determinant = -1

    def __str__(self):
        return 'SymOp.Inversion {} <{}>'.format(self._label, hex(id(self)))

    def get_measure_modes(self, modes, orientation=None):

        operation = inversion()

        measure_mode = []
        for mode in modes:
            operated_mode = np.dot(operation, np.array(mode).T).T
            norm = np.linalg.norm(mode)
            permu_mode = np.array(operated_mode)[self.permutation]

            measure_mode.append(np.trace(np.dot(mode, permu_mode.T))/norm)

        return np.array(measure_mode)

    def get_measure_atom(self):

        measure_atoms = np.array([1 if i == p else 0 for i, p in enumerate(self.permutation)])

        return np.sum(measure_atoms)

    def get_measure_xyz(self, orientation=None):

        operation = inversion()

        measure_mode = []
        for axis in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
            operated_axis = np.dot(operation, axis)
            measure_mode.append(np.dot(axis, operated_axis))

        return np.sum(measure_mode)

    def get_displacements_projection(self, orientation=None):

        n_atoms = len(self.permutation)
        operation = inversion()

        projected_modes = []
        cartesian_modes = np.identity(3 * n_atoms).reshape(3 * n_atoms, n_atoms, 3)

        for i, mode in enumerate(cartesian_modes):
            operated_mode = np.dot(operation, np.array(mode).T).T
            projected_modes.append(operated_mode[self.permutation])

        return np.array(projected_modes)

    def get_measure_pos(self, coordinates, orientation=None, normalized=True):

        operation = inversion()
        operated_coor = np.dot(operation, coordinates.T).T
        mesure_pos = np.einsum('ij, ij -> ', coordinates, operated_coor[self.permutation])

        if normalized:
            mesure_pos /= np.einsum('ij, ij -> ', coordinates, coordinates)

        return mesure_pos

    def get_operated_coordinates(self, coordinates, orientation=None):

        operation = inversion()
        operated_coor = np.dot(operation, coordinates.T).T

        return operated_coor[self.permutation]

    def get_overlap_func(self, op_function1, op_function2, orientation=None):

        operation = inversion()

        op_function_i = op_function1.copy()
        op_function_i.apply_linear_transformation(operation)

        return (op_function2*op_function_i).integrate

    def apply_rotation(self, orientation):
        pass

    @property
    def matrix_representation(self):
        return inversion()



from posym.operations import Operation
from scipy.spatial.transform import Rotation as R
from posym.tools import standardize_vector
import numpy as np


def cache_rotation(func):
    cache_dict = {}

    def wrapper_cache(angle, rotation_axis):
        hash_key = (angle, tuple(rotation_axis))
        if hash_key in cache_dict:
            return cache_dict[hash_key]

        cache_dict[hash_key] = func(angle, rotation_axis)
        return cache_dict[hash_key]

    return wrapper_cache

@cache_rotation
def rotation(angle, rotation_axis):

    rotation_vector = angle * np.array(rotation_axis) / np.linalg.norm(rotation_axis)
    rotation = R.from_rotvec(rotation_vector)

    return rotation.as_matrix()


def prepare_vector(positions, vector):
    return np.array(vector) + np.array(positions)


class Rotation(Operation):
    def __init__(self, label, axis, order=1, exp=1):
        super().__init__(label)

        self._axis = standardize_vector(axis)
        self._order = order
        self._exp = np.mod(exp, order)
        self._determinant = 1

        # normalize C2
        if order <= 2:
            self._exp = abs(exp)

        if self._order > 10:
            print('ko', self._order)
            exit()

    def __str__(self):
        axis_txt = '[{:8.3f} {:8.3f} {:8.3f}]'.format(*self._axis)
        return 'SymOp.Rotation {} {} order: {} exp: {} <{}>'.format(self._label, axis_txt, self._order, self._exp, hex(id(self)))

    def get_measure_modes(self, modes, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        measure_mode = []
        for mode in modes:
            operated_mode = np.dot(operation, np.array(mode).T).T
            norm = np.linalg.norm(mode)
            permu_mode = np.array(operated_mode)[self.permutation]
            measure_mode.append(np.trace(np.dot(mode, permu_mode.T))/norm)

        return measure_mode

    def get_measure_atom(self):

        measure_atoms = np.array([1 if i == p else 0 for i, p in enumerate(self.permutation)])

        return np.sum(measure_atoms)

    def get_measure_xyz(self, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        measure_mode = []
        for axis in [[1, 0, 0], [0, 1, 0], [0, 0, 1]]:
            operated_axis = np.dot(operation, axis)
            measure_mode.append(np.dot(axis, operated_axis))

        return np.sum(measure_mode)

    def get_displacements_projection(self, orientation=None):

        n_atoms = len(self.permutation)
        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        cartesian_modes = np.identity(3 * n_atoms).reshape(3 * n_atoms, n_atoms, 3)
        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        projected_modes = []
        for i, mode in enumerate(cartesian_modes):
            operated_mode = np.dot(operation, np.array(mode).T).T
            projected_modes.append(operated_mode[self.permutation])

        return projected_modes

    def get_measure_pos(self, coordinates, orientation=None, normalized=True):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = rotation(2 * np.pi / self._order * self._exp, rotated_axis)

        operated_coor = np.dot(operation, coordinates.T).T
        mesure_pos = np.einsum('ij, ij -> ', coordinates, operated_coor[self.permutation])

        if normalized:
            mesure_pos /= np.einsum('ij, ij -> ', coordinates, coordinates)

        return mesure_pos

    def get_operated_coordinates(self, coordinates, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)
        operation = rotation(2 * np.pi / self._order * self._exp, rotated_axis)

        operated_coor = np.dot(operation, coordinates.T).T

        return operated_coor[self.permutation]

    def get_overlap_func(self, op_function1, op_function2, orientation=None):

        rotated_axis = self._axis if orientation is None else orientation.apply(self._axis)

        angle = 2 * np.pi / self._order * self._exp
        operation = rotation(angle, rotated_axis)

        fn_function_r = op_function1.copy()
        fn_function_r.apply_linear_transformation(operation)

        return (op_function2*fn_function_r).integrate

    def apply_rotation(self, orientation):
        self._axis = orientation.apply(self._axis)

    def inverse(self):
        return Rotation(self.label, axis=self._axis, order=self._order, exp=-self._exp)

    @property
    def axis(self):
        return self._axis

    @property
    def order(self):
        return self._order

    @property
    def exp(self):
        return self._exp

    @property
    def matrix_representation(self):
        angle = 2 * np.pi / self._order * self._exp
        return rotation(angle, self._axis)

